# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from scipy.special import erf

from quocslib.pulses.BasePulse import BasePulse
from quocslib.pulses.basis.ChoppedBasis import ChoppedBasis
from quocslib.tools.randomgenerator import RandomNumberGenerator


class SigmoidAllParas(ChoppedBasis):
    amplitude_variation: float
    optimized_control_parameters: np.ndarray
    optimized_super_parameters: np.ndarray
    time_grid: np.ndarray

    def __init__(self, map_index: int, pulse_dictionary: dict, rng: RandomNumberGenerator = None, is_AD: bool = False):
        """
        Constructor of the Sigmoid basis class with adjustable superparameters. It calls the constructor of the parent
        class ChoppedBasis.

        :param int map_index: Index number to use to get the control parameter.
        :param dict pulse_dictionary: The dictionary of the pulse is defined here. Only the basis dictionary is used.
        """
        basis_dict = pulse_dictionary["basis"]

        # Frequencies number i.e. the basis vector number in the pulse parametrization
        self.super_parameter_number = basis_dict.setdefault("basis_vector_number", 1)

        # define offset from side and sigma
        self.sigma = basis_dict.setdefault("sigma", 10**(-10))
        self.offset = basis_dict.setdefault("offset", -10*self.sigma)
        # tau initial variation
        self.tau_var = basis_dict.setdefault("tau_variation", 10**(-20))
        # sigma initial variation
        self.sigma_var = basis_dict.setdefault("sigma_variation", 10**(-20))

        # Number of control parameters to be optimized
        self.control_parameters_number = self.super_parameter_number + 1
        if self.tau_var > 10**(-10):
            self.control_parameters_number += self.super_parameter_number
        if self.sigma_var > 10**(-10):
            self.control_parameters_number += self.super_parameter_number

        # Constructor of the parent class Chopped Basis
        super().__init__(map_index=map_index, rng=rng, is_AD=is_AD, **pulse_dictionary)

        # Define scale and offset coefficients
        self.scale_coefficients = self.amplitude_variation * np.ones((self.control_parameters_number, ))
        self.offset_coefficients = np.zeros((self.control_parameters_number, ))

    def _get_shaped_pulse(self) -> np.array:
        """
        Definition of the pulse parametrization. It is called at every function evaluation to build the pulse and
        return it as an array.

        :return np.array: The pulse as an array.
        """
        final_time = self.final_time
        tauN = len(self.super_parameter_distribution_obj.w)

        # properly distributing the optimised parameters
        if self.tau_var > 10**(-10):
            if self.sigma_var > 10**(-10):
                Aopti = self.optimized_control_parameters[0:tauN+1]
                # time step adjustments
                del_tau = self.optimized_control_parameters[tauN+1:2*tauN+1]\
                          * self.tau_var / self.amplitude_variation
                # sigma adjustments
                del_sig = abs(self.optimized_control_parameters[2*tauN+1:]) \
                          * self.sigma_var / self.amplitude_variation
            else:
                Aopti = self.optimized_control_parameters[0:tauN+1]
                # time step adjustments
                del_tau = self.optimized_control_parameters[tauN+1:]\
                          * self.tau_var / self.amplitude_variation
                # sigma adjustments
                del_sig = np.zeros(tauN)
        elif self.sigma_var > 10**(-10):
            Aopti = self.optimized_control_parameters[0:tauN+1]
            # time step adjustments
            del_tau = np.zeros(tauN)
            # sigma adjustments
            del_sig = abs(self.optimized_control_parameters[tauN+1:]) \
                      * self.sigma_var / self.amplitude_variation
        else:
            Aopti = self.optimized_control_parameters
            # time step adjustments
            del_tau = np.zeros(tauN)
            # sigma adjustments
            del_sig = np.zeros(tauN)

        taus = np.zeros(tauN)
        taus += self.super_parameter_distribution_obj.w
        if self.tau_var > 10**(-10):
            # adjusting the taus
            taus += del_tau
            sorted_tau = np.sort(taus)
            # making sure the time steps don't move outside of the pulse
            if len(sorted_tau) == 1:
                if sorted_tau[0] < self.offset:
                    taus[0] = self.offset
                elif sorted_tau[0] > final_time - self.offset:
                    taus[0] = final_time - self.offset
            elif sorted_tau[0] < self.offset or sorted_tau[-1] > final_time - self.offset:
                if sorted_tau[-1] < final_time - self.offset:
                    sorted_tau[-1] = final_time - self.offset
                if sorted_tau[0] > self.offset:
                    sorted_tau[0] = self.offset
                rel_tau = taus - (sorted_tau[-1]-sorted_tau[0])/2 - sorted_tau[0]
                taus = final_time/2 + rel_tau * 2 * (final_time/2 - self.offset)/(sorted_tau[-1]-sorted_tau[0])

        sigmas = del_sig + self.sigma
        # adjusting the sigmas
        if (self.sigma_var > 10**(-10)) & (self.offset >= 10**(-10)):
            # making sure it stays in range if the offset is unequal zero
            kappa = self.sigma/self.offset
            for i in range(len(sigmas)):
                del_sigma_max = (final_time/2 - abs(final_time/2-taus[i])) * kappa - self.sigma
                sigmas[i] = self.sigma + del_sigma_max*(1-np.cos(np.pi*del_sig[i]))/2

        # building the pulse
        t = self.time_grid
        pulse = np.zeros(self.bins_number)
        for ii in range(self.super_parameter_number):
            pulse += (Aopti[ii + 1] / 2 * (erf((t - taus[ii]) / (np.sqrt(2) * sigmas[ii])) + 1))
        pulse += Aopti[0] / 2 * (erf((t - self.offset) / (np.sqrt(2) * self.sigma)) + 1)
        if self.offset >= 10**(-10):
            pulse += (-np.sum(Aopti) / 2 * (erf((t - (final_time - self.offset)) / (np.sqrt(2) * self.sigma)) + 1))
        return pulse



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import os
import time
import numpy as np
import json

from quocslib.utils.AbstractDump import AbstractDump


class NumpyEncoder(json.JSONEncoder):
    """
    Class to convert numpy arrays to lists for json dumping
    """
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.float128):
            # for some reason float128 can lead to the error:
            # TypeError: Object of type float128 is not JSON serializable
            return np.float64(obj)
        return json.JSONEncoder.default(self, obj)


class BestDump(AbstractDump):
    """
    Dumping class for controls and other data which should be the most useful option for most users.
    """

    def __init__(self, results_path: str = ".", date_time: str = ".", dump_format: str = "npz", **kwargs):
        """
        Constructor for the BestDump class.

        :param str results_path: Path of the folder of the results
        :param str date_time: String containing the identifier in the form of date and time
        :param str dump_format: Format of the dump, either "npz" or "json"
        """
        self.best_controls_path = results_path
        self.results_path = results_path
        self.date_time = date_time
        self.dump_format = dump_format

    def dump_controls(self,
                      pulses: list = [],
                      timegrids: list = [],
                      parameters: list = [],
                      is_record: bool = False,
                      **kwargs) -> None:
        """
        Saves the controls in the results folder.

        :param list pulses: the list containing the pulses that were optimized
        :param list timegrids: the list containing the time grids that were used in the optimization
        :param list parameters: the list containing the parameters that were optimized
        :param bool is_record: information if the current controls are a new record
        """
        if not is_record:
            return

        controls_dict = {}
        pulse_index = 0
        pulse_names = []
        if "pulse_names" in {**kwargs}:
            pulse_names = {**kwargs}["pulse_names"]
        else:
            pulse_names = ["pulse_{}".format(index+1) for index in range(len(pulses))]

        # time_names = []
        # if "time_names" in {**kwargs}:
        #     time_names = {**kwargs}["time_names"]
        # else:
        #     time_names = ["time_grid_{}".format(index + 1) for index in range(len(timegrids))]

        for pulse, time_grid in zip(pulses, timegrids):
            pulse_name = pulse_names[pulse_index]
            if pulse_name in controls_dict:
                pulse_name = pulse_name + str(pulse_index + 1)
            time_name = "time_grid_for_" + pulse_name
            controls_dict[pulse_name] = pulse
            controls_dict[time_name] = time_grid
            pulse_index += 1

        parameter_index = 0
        parameter_names = []
        if "parameter_names" in {**kwargs}:
            parameter_names = {**kwargs}["parameter_names"]
        else:
            parameter_names = ["parameter_{}".format(index+1) for index in range(len(parameters))]

        for parameter in parameters:
            param_name = parameter_names[parameter_index]
            if param_name in controls_dict:
                param_name = param_name + str(parameter_index + 1)
            controls_dict[param_name] = parameter
            parameter_index += 1

        # Full dictionary
        full_dict = {**controls_dict, **kwargs}

        # Save the file
        if self.dump_format == "json":
            self.dump_dict("best_controls", full_dict)
        else:
            controls_path = os.path.join(self.results_path, self.date_time + "_best_controls.npz")
            np.savez(controls_path, **full_dict)

        # if "iteration_number" in full_dict:
        #     iteration_path = os.path.join(self.results_path, self.date_time + "_funct_eval_of_best_controls.txt")
        #     with open(iteration_path, 'w') as f:
        #         f.write(str(full_dict["iteration_number"]))

    def other_dumps(self, filename: str = "test.txt", data: np.array = np.array([0.0])):
        """
        Saves other results into a txt numpy file.

        :param str filename: name of the file
        :param np.array data: data to be saved
        """
        # Create the path
        path = os.path.join(self.results_path, filename)
        # Save the data in a txt file
        np.savetxt(path, data)

    def dump_dict(self, data_file_name: str = "unknown_data_dict", data_dict=None):
        """
        Saves a dictionary to a file.

        :param str data_file_name: name of the file
        :param dict data_dict: dictionary to be saved
        """
        if data_dict is None:
            data_dict = {}
        data_dict_path = os.path.join(self.results_path, self.date_time + "_" + data_file_name+'.json')
        with open(data_dict_path, 'w') as convert_file:
            convert_file.write(json.dumps(data_dict, indent=4, cls=NumpyEncoder))




# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.optimizationalgorithms.OptimizationAlgorithm import OptimizationAlgorithm
from quocslib.Controls import Controls
from quocslib.gradientfreemethods.NelderMead import NelderMead
from quocslib.utils.dynamicimport import dynamic_import
from quocslib.tools.linearalgebra import simplex_creation
import numpy as np
from scipy.stats import norm


class DirectSearchAlgorithm(OptimizationAlgorithm):
    """
    This is an implementation of a direct search algorithm. It only performs a direct search based on the specified
    search method in the dsm_settings of the optimization_dict. The algorithm is stopped when the stopping criteria
    are met. The algorithm is implemented in the run function. The results are returned in the _get_final_results
    function.
    """

    initStatus = 0
    terminate_reason = "-1"

    def __init__(self, optimization_dict: dict = None, communication_obj=None, **kwargs):
        """
        Constructor for the direct search algorithm. The optimization dictionary is passed to the constructor of the
        OptimizationAlgorithm class. The direct search method is initialized here.
        :param optimization_dict: dictionary with the optimization settings
        :param communication_obj: communication object to communicate with the client
        """
        super().__init__(communication_obj=communication_obj, optimization_dict=optimization_dict)
        ###########################################################################################
        # Direct Search method
        ###########################################################################################
        stopping_criteria = optimization_dict["algorithm_settings"]["dsm_settings"]["stopping_criteria"]
        # put global time limit into stopping_criteria, so we don't have to pass it through functions
        optimization_dict["algorithm_settings"].setdefault("total_time_lim", 10**10)
        stopping_criteria.setdefault("total_time_lim", optimization_dict["algorithm_settings"]["total_time_lim"])
        direct_search_method_settings = optimization_dict["algorithm_settings"]["dsm_settings"]["general_settings"]
        if "dsm_name" in direct_search_method_settings:
            print("dsm_name is used direct search methods. This option is deprecated. Use \n"
                  "dsm_algorithm_module: quocslib.freegradients.NelderMead\n"
                  "dsm_algorithm_class: NelderMead")
            self.dsm_obj = NelderMead(direct_search_method_settings,
                                      stopping_criteria,
                                      callback=self.is_optimization_running)
        else:
            dsm_attribute = dynamic_import(
                module_name=direct_search_method_settings.setdefault("dsm_algorithm_module", None),
                class_name=direct_search_method_settings.setdefault("dsm_algorithm_class", None),
                name=direct_search_method_settings.setdefault("dsm_algorithm_name", None),
                class_type='dsm_settings')

            self.dsm_obj = dsm_attribute(direct_search_method_settings,
                                         stopping_criteria,
                                         callback=self.is_optimization_running)

        ################################################################################################
        # Extract some Settings for the Search #########################################################
        ################################################################################################
        alg_parameters = optimization_dict["algorithm_settings"]
        drift_compensations_parameters = optimization_dict["algorithm_settings"].setdefault("compensate_drift", {})
        # Define sigma for re-evaluation steps
        self.best_sigma = 0.0
        # Update the FoM for drift
        self.compensate_drift_after_minutes = drift_compensations_parameters.setdefault("compensate_after_minutes", 0.0)
        self.compensate_drift_num_average = drift_compensations_parameters.setdefault("num_average", 1)
        # Re-evaluation steps option
        if "re_evaluation" in alg_parameters:
            re_evaluation_parameters = alg_parameters["re_evaluation"]
            if "re_evaluation_steps" in re_evaluation_parameters:
                self.re_evaluation_steps = np.asarray(re_evaluation_parameters["re_evaluation_steps"], dtype=float)
            else:
                self.re_evaluation_steps = np.asarray([0.3, 0.5, 0.501], dtype=float)
                message = "Steps not found. The default will be used in the optimization: {0}".format(
                    self.re_evaluation_steps)
                self.comm_obj.print_logger(message, level=30)
            # Define the FoM test and sigma test arrays, with the maximum number of steps
            # FoM test and sigma test are arrays containing the FoM and sigma at every re-evaluation step
            self.FoM_test = np.zeros(self.re_evaluation_steps.shape[0] + 1, dtype=float)
            self.sigma_test = np.zeros_like(self.FoM_test)
        else:
            self.re_evaluation_steps = None

        ###########################################################################################
        # Optimal algorithm variables ###########################################################
        ###########################################################################################
        # Starting FoM
        # self.best_FoM = 1e10  # defined in parent class
        ###########################################################################################
        # Pulses, Parameters object ###########################################################
        ###########################################################################################
        self.controls = Controls(optimization_dict["pulses"], optimization_dict["times"],
                                 optimization_dict["parameters"])
        self.FoM_list: list = []
        self.iteration_number_list: list = []

    def _inner_routine_call(self, optimized_control_parameters: np.array, iterations: int,
                            drift_comp_new_val=None) -> float:
        """
        Wrapper Function for the _routine_call of the OptimizationAlgorithm class.
        Used to perform drift compensation and re-evaluation steps.
        :param optimized_control_parameters: optimized control parameters
        :param iterations: number of iterations
        :param drift_comp_new_val: new FoM value in case a drift compensation happened
        :return float: FoM
        """
        if drift_comp_new_val is not None:
            mu_1 = drift_comp_new_val
            self.best_FoM = mu_1
            self.is_record = True
            message = "New record due to drift compensation. New best FoM: {0}".format(self.FoM_factor*mu_1)
            self.comm_obj.print_logger(message, level=20)
            self.best_xx = self.xx.copy()
            self.comm_obj.update_controls(is_record=True,
                                          FoM=self.best_FoM,
                                          # super_it=self.super_it,
                                          iteration_number=self.iteration_number)
        else:
            # set the is_record to False unless we have drift_comp call... then we want to update the current best FoM
            self.is_record = False
            # Initialize step number to 0
            self.step_number = 0
            FoM = self._routine_call(optimized_control_parameters, iterations)

            ################################################################################################################
            # Standard function evaluation - dCRAB without re-evaluation steps
            ################################################################################################################
            if self.re_evaluation_steps is None:
                mu_1 = FoM
            else:
                ############################################################################################################
                # Implement the re-evaluation step method
                ############################################################################################################
                # check mu-sig criterion by calculating probability of current pulses being new record
                # Re evaluation steps initialization e.g. [0.33, 0.5, 0.501, 0.51]
                re_evaluation_steps = self.re_evaluation_steps
                # First evaluation in whole optimization -> do not reevaluate
                if self.iteration_number == 1:
                    re_evaluation_steps = np.array([0.5])
                # number of steps
                max_steps_number = re_evaluation_steps.shape[0]
                # Initialize to zero the FoM_test and the sigma_test arrays
                self.FoM_test = 0.0 * self.FoM_test
                self.sigma_test = 0.0 * self.sigma_test
                # Get the figure of merit from the client
                self.FoM_test[0] = FoM
                # TODO: Check if optimization_is_running is necessary here
                # Get the standard deviation
                self.sigma_test[0] = float(self.FoM_dict.setdefault("std", 1.0))
                # Increase step number after function evaluation
                self.step_number += 1
                # p level test better than current record
                for ii in range(max_steps_number):
                    p_level = re_evaluation_steps[ii]
                    mu_1, sigma_1 = self._get_average_FoM_std(mu_sum=np.sum(self.FoM_test) * 1.0,
                                                              sigma_sum=np.sum(self.sigma_test) * 1.0)
                    mu_2, sigma_2 = self.best_FoM, self.best_sigma
                    probability = self._probabnormx1betterx2(mu_1, sigma_1, mu_2, sigma_2)
                    # If probability is lower than the probability in the list return the
                    if probability < p_level:
                        return mu_1
                    # else: go on with further re-evaluations
                    self.FoM_test[ii + 1] = self._routine_call(optimized_control_parameters, iterations)
                    self.sigma_test[ii + 1] = float(self.FoM_dict.setdefault("std", 1.0))
                    # Increase step number after function evaluation
                    self.step_number += 1

                # check if last threshold (re_evaluation_steps[-1]) is surpassed -> new record
                mu_1, sigma_1 = self._get_average_FoM_std(mu_sum=np.sum(self.FoM_test) * 1.0,
                                                          sigma_sum=np.sum(self.sigma_test) * 1.0)
                mu_2, sigma_2 = self.best_FoM, self.best_sigma
                probability = self._probabnormx1betterx2(mu_1, sigma_1, mu_2, sigma_2)
                # TODO: Check what best FoM means in this case
                if probability > re_evaluation_steps[-1]:
                    # We have a new record
                    self.best_sigma, self.best_FoM = sigma_1, mu_1
                    self.is_record = True
                    message = "New record achieved. New best FoM: {0}, std: {1}".format(self.FoM_factor*mu_1, sigma_1)
                    self.comm_obj.print_logger(message, level=20)
                    self.best_xx = self.xx.copy()
                    self.comm_obj.update_controls(is_record=True,
                                                  FoM=self.best_FoM,
                                                  sigma=self.best_sigma,
                                                  # super_it=self.super_it,
                                                  iteration_number=self.iteration_number)

        # Return the figure of merit to be minimized by the updating algorithm
        return mu_1

    def _get_average_FoM_std(self, mu_sum: float = None, sigma_sum: float = None) -> np.array:
        """
        Calculate the average figure of merit and standard deviation based on the passed arrays

        :param mu_sum: Sum of the figure of merit values
        :param sigma_sum: Sum of the standard deviation values
        :return float, float: Average figure of merit and standard deviation
        """
        step_number = self.step_number
        # For the first evaluation and in case no re-evaluation step is needed return directly
        if step_number == 0:
            return self.FoM_dict["FoM"], self.FoM_dict.setdefault("std", 1.0)
        # Call from the response for client function. Calculate the average FoM based on all the previous FoM stored in
        # the FoM_test array and the current FoM in the FoM dict
        if mu_sum is None:
            curr_FoM, curr_std = self.FoM_dict["FoM"], self.FoM_dict.setdefault("std", 1.0)
            mu_sum = np.mean(self.FoM_test[:step_number]) * (step_number - 1) + curr_FoM
            sigma_sum = (np.mean(self.sigma_test[:step_number]) * (step_number - 1) + curr_std)
        # If it is called inside the _inner_routine_call()
        average_FoM, average_std = mu_sum / step_number, sigma_sum / step_number
        return average_FoM, average_std

    def _probabnormx1betterx2(self, mu_1: float, sigma_1: float, mu_2: float, sigma_2: float):
        """
        Calculates probability for normal distributed random variable x1 being greater or equal than x2
        x1 usually refers to the test pulse and
        x2 to the current record that is tried to be outperformed

        :param mu_1: = <x1>
        :param sigma_1: = std(x1)
        :param mu_2:  = <x2>
        :param sigma_2: = std(x2)
        :return float: probability P(x1>=x2)
        """
        # Start by defining a new random variable z = x1 - x2
        # if mu_z > 0 the probability is > 0.5 , else: <0.5
        mu_z = mu_2 - mu_1
        std_comb = np.sqrt(sigma_1**2 + sigma_2**2)
        if np.abs(std_comb) < 10**(-14):
            # Warning message
            message = ("Combined standard deviation std_comb = {0} < 10**(-14) . To avoid numerical instabilities "
                       "std_comb will be set equal to 1.0".format(std_comb))
            self.comm_obj.print_logger(message, level=30)
            # Set std_com to 1.0
            std_comb = 1.0
        zz = mu_z / std_comb
        # Calculate the probability with the cumulative density function
        probability = norm.cdf(zz)
        return probability

    def _get_response_for_client(self):
        """
        Return useful information for the client interface and print message in the log

        :return dict: Dictionary containing useful information for the client interface
        """
        FoM = self.FoM_dict["FoM"]
        status_code = self.FoM_dict.setdefault("status_code", 0)
        # If re-evaluation steps is not used check for current best figure of merit
        if self.re_evaluation_steps is None:
            if self.get_is_record(FoM):
                message = "New record achieved. Previous FoM: {FoM}, new best FoM : {best_FoM}".format(
                    FoM=self.FoM_factor*self.best_FoM, best_FoM=self.FoM_factor*FoM)
                self.comm_obj.print_logger(message=message, level=20)
                self.best_FoM = FoM
                self.best_xx = self.xx.copy()
                self.is_record = True
        response_dict = {
            "is_record": self.is_record,
            "FoM": FoM,
            "iteration_number": self.iteration_number,
            "status_code": status_code
        }

        if self.re_evaluation_steps is not None:
            FoM, std = self._get_average_FoM_std()
            message = "Function evaluation number: {func_eval}, ".format(func_eval=self.iteration_number)
            message += "Re-eval. number: {0}, FoM: {1}, std: {2}".format(self.step_number, self.FoM_factor*FoM, std)
            self.comm_obj.print_logger(message, level=20)

        if status_code == 0:
            self.FoM_list.append(self.FoM_factor*FoM)
            self.iteration_number_list.append(self.iteration_number)
        return response_dict

    def run(self):
        """
        Runs the optimization algorithm
        """
        # Direct search method
        self._dsm_build()

    def _dsm_build(self):
        """
        Runs the direct search
        """
        start_simplex = simplex_creation(self.controls.get_mean_value(), self.controls.get_sigma_variation())
        # Initial point for the Start Simplex
        x0 = self.controls.get_mean_value()
        # Initialize the best xx vector for this SI
        self.best_xx = self.controls.get_mean_value().copy()
        # Run the direct search algorithm
        result_l = self.dsm_obj.run_dsm(self._inner_routine_call,
                                        x0,
                                        initial_simplex=start_simplex,
                                        sigma_v=self.controls.get_sigma_variation(),
                                        drift_comp_minutes=self.compensate_drift_after_minutes,
                                        drift_comp_num_average=self.compensate_drift_num_average)
        # Update the results
        [FoM, xx,
         self.terminate_reason] = [result_l["F_min_val"], result_l["X_opti_vec"], result_l["terminate_reason"]]

    def _get_controls(self, xx):
        """
        Returns the controls dictionary for the given input vector
        :param xx:
        :return dict: Dictionary containing the controls
        """
        # pulses_list, time_grids_list, parameters_list
        [], [], parameters_list = self.controls.get_controls_lists(xx)

        controls_dict = {"pulses": [], "parameters": parameters_list, "timegrids": []}
        return controls_dict

    def _get_final_results(self):
        """
        Returns the final results of the optimization

        :return dict: Dictionary containing the final results
        """
        final_dict = {
            "Figure of merit": self.FoM_factor * self.best_FoM,
            "parameters": self.xx,
            "terminate_reason": self.terminate_reason
        }
        return final_dict



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np

from quocslib.pulses.basis.ChoppedBasis import ChoppedBasis
from quocslib.parameters.Parameter import Parameter
from quocslib.parameters.TimeParameter import TimeParameter
from quocslib.tools.randomgenerator import RandomNumberGenerator
from quocslib.utils.dynamicimport import dynamic_import

from typing import List, Tuple


class Controls:
    """
    This is the main class for the optimization quantities, i.e. pulses, parameters, and times.
    All these quantities are defined in this class and can be accessed by calling the modules here.
    """

    def __init__(self, pulses_list, times_list, parameters_list, rng: RandomNumberGenerator = None,
                 is_AD: bool = False):
        """
        Constructor of the general class containing all the controls used during the optimization

        :param pulses_list: List containing the settings for each pulse
        :param times_list: List containing the settings for each time
        :param parameters_list: List containing the setting for each parameter
        :param rng: Random number generator object (optional)
        :param is_AD: Boolean flag to activate the automatic differentiation (optional)
        """
        # Map index
        map_index = -1
        # Lists and dictionary for the controls
        self.pulse_objs_list = []
        self.parameter_objs_list = []
        self.times_obj_dictionary = {}
        ###############################################
        # Pulses
        ###############################################
        self.pulses_number = len(pulses_list)
        for pulse in pulses_list:
            # Basis attribute inside the dictionary or module and class relative import
            basis = pulse["basis"]
            basis_attribute = dynamic_import(attribute=basis.setdefault("basis_attribute", None),
                                             module_name=basis.setdefault("basis_module", None),
                                             class_name=basis.setdefault("basis_class", None),
                                             name=basis.setdefault("basis_name", None),
                                             class_type='basis')
            # Create the pulse obj
            self.pulse_objs_list.append(basis_attribute(map_index, pulse, rng=rng, is_AD=is_AD))
            # Update the map index for the next control
            map_index = self.pulse_objs_list[-1].last_index
        ###############################################
        # Parameters
        ###############################################
        self.parameters_number = len(parameters_list)
        for parameter in parameters_list:
            self.parameter_objs_list.append(Parameter(map_index, parameter, is_AD=is_AD))
            # Update the map index for the next control
            map_index = self.parameter_objs_list[-1].last_index
        ###############################################
        # Times
        ###############################################
        self.times_number = len(times_list)
        for time in times_list:
            self.times_obj_dictionary[time["time_name"]] = TimeParameter(**time)
            # TODO Implement the time optimization here

        # Custom get control functions for AD
        if is_AD:
            self.get_controls_lists = self._get_controls_jax_obj
            self.get_final_controls_lists = self._get_converted_jax_obj
        else:
            self.get_controls_lists = self._get_controls_lists
            self.get_final_controls_lists = self._get_controls_lists
        # Calculate the biggest bins number among the different pulses
        if len(self.pulse_objs_list) > 0:
            self.max_bin_numbers = np.max(np.asarray([pulse.bins_number for pulse in self.pulse_objs_list], dtype=int))
        else:
            self.max_bin_numbers = 0
        # Load the AD functions
        if is_AD:
            self._buffer_AD_arrays()

    def _buffer_AD_arrays(self):
        """
        Imports jax.numpy if AD is activated
        """
        import jax.numpy as jnp
        import jax
        # self.debug_print = jax.debug.print
        self.jnp = jnp

    def get_control_parameters_number(self) -> int:
        """Returns the number of control parameters"""
        ###############################################
        # Pulses
        ###############################################
        control_parameters_number = 0
        for pulse_obj in self.pulse_objs_list:
            # Update number control parameters
            control_parameters_number += pulse_obj.control_parameters_number
        ###############################################
        # Parameters
        ###############################################
        for parameter_obj in self.parameter_objs_list:
            # Update number control parameters
            control_parameters_number += parameter_obj.control_parameters_number
        ###############################################
        # Times
        ###############################################
        # TODO Implement the time optimization here
        # Return the control parameters number
        return control_parameters_number

    def select_basis(self) -> None:
        """Initializes the super parameter basis"""
        for pulse in self.pulse_objs_list:
            if isinstance(pulse, ChoppedBasis):
                pulse.super_parameter_distribution_obj.set_random_super_parameter()
                # Update the base pulse parameters and functions
                pulse.update_chopped_basis()
                # Update the control parameter indexes
                self._update_control_parameter_indexes()

    def _update_control_parameter_indexes(self) -> None:
        """Updates the control parameter indexes"""
        # Map index
        map_index = -1
        ###############################################
        # Pulses
        ###############################################
        for pulse_obj in self.pulse_objs_list:
            pulse_obj.set_control_parameters_list(map_index)
            # Update the map index for the next control
            map_index = pulse_obj.last_index
        ###############################################
        # Parameters
        ###############################################
        for parameter_obj in self.parameter_objs_list:
            parameter_obj.set_control_parameters_list(map_index)
            # Update the map index for the next control
            map_index = parameter_obj.last_index
        ###############################################
        # Times
        ###############################################
        # TODO Implement the time optimization here
        # Return the control parameters number

    def get_random_super_parameter(self) -> List[float]:
        """
        Returns a list with current super_parameters

        :return list: List with current super_parameters
        """
        super_parameter_list = []
        for pulse in self.pulse_objs_list:
            if isinstance(pulse, ChoppedBasis):
                super_parameter_list.append(list(pulse.super_parameter_distribution_obj.w))
        # super_parameter_array = np.asarray(super_parameter_list)
        return super_parameter_list

    def get_sigma_variation(self) -> np.array:
        """
        Returns a numpy array with the maximum sigma in the parameters choice for the start simplex

        :return np.array:
        """
        sigma_variation_coefficients = np.zeros(self.get_control_parameters_number(), dtype="float")
        # Pulses
        for pulse in self.pulse_objs_list:
            sigma_variation_coefficients[pulse.control_parameters_list] = pulse.scale_coefficients
        # Parameters
        for parameter in self.parameter_objs_list:
            sigma_variation_coefficients[parameter.control_parameters_list] = parameter.amplitude_variation
        # Times
        for time_name in self.times_obj_dictionary:
            time = self.times_obj_dictionary[time_name]
            if time.is_optimization:
                sigma_variation_coefficients[time.control_parameters_list] = time.amplitude_variation
        return sigma_variation_coefficients

    def get_mean_value(self) -> np.array:
        """
        Return a numpy array with the mean value of the control parameters

        :return np.array:
        """
        mean_value_coefficients = np.zeros(self.get_control_parameters_number(), dtype="float")
        # Pulses
        for pulse in self.pulse_objs_list:
            mean_value_coefficients[pulse.control_parameters_list] = pulse.offset_coefficients
        # Parameters
        for parameter in self.parameter_objs_list:
            mean_value_coefficients[parameter.control_parameters_list] = parameter.value
        # Times
        # TODO In case of time optimization return the initial value
        return mean_value_coefficients

    def update_base_controls(self, optimized_parameters_vector: np.array) -> None:
        """
        Updates the base controls with the optimized parameters.

        :param np.array optimized_parameters_vector: Optimized parameters vector
        """
        # Set the times
        for time_name in self.times_obj_dictionary:
            time = self.times_obj_dictionary[time_name]
            if time.is_optimization:
                time.set_parameter(optimized_parameters_vector[time.control_parameters_list])
        # Set the pulses
        for pulse in self.pulse_objs_list:
            time_name = pulse.time_name
            pulse.set_base_pulse(optimized_parameters_vector[pulse.control_parameters_list],
                                 final_time=self.times_obj_dictionary[time_name].get_time())
        # Set the parameters
        for parameter in self.parameter_objs_list:
            parameter.set_parameter(optimized_parameters_vector[parameter.control_parameters_list])

    def _get_controls_lists(self, optimized_parameters_vector: np.array) -> Tuple[List[float], List[float], List[float]]:
        """
        Sets the optimized control parameters and get the controls

        :param np.array optimized_parameters_vector:
        :return: The pulses, time grids, and the parameters in three different lists of numpy arrays.
        """
        pulses_list = []
        time_grids_list = []
        parameters_list = []
        # Set the times
        for time_name in self.times_obj_dictionary:
            time = self.times_obj_dictionary[time_name]
            if time.is_optimization:
                time.set_parameter(optimized_parameters_vector[time.control_parameters_list])
        # Get the pulses and the timegrids
        for pulse in self.pulse_objs_list:
            time_name = pulse.time_name
            pulses_list.append(
                pulse.get_pulse(optimized_parameters_vector[pulse.control_parameters_list],
                                final_time=self.times_obj_dictionary[time_name].get_time()))
            time_grids_list.append(pulse.time_grid)
        # Get the parameters
        for parameter in self.parameter_objs_list:
            parameters_list.append(
                parameter.get_parameter(optimized_parameters_vector[parameter.control_parameters_list]))

        return pulses_list, time_grids_list, parameters_list

    def _get_converted_jax_obj(self, optimized_parameters_vector: np.array) -> np.array:
        """
        Converts the jax controls into numpy controls casting to real

        :param np.array optimized_parameters_vector:
        :return: The pulses, time grids, and the parameters in three different lists of numpy arrays.
        """
        pulses_array = self.jnp.zeros((len(self.pulse_objs_list), self.max_bin_numbers))
        time_grids_array = self.jnp.zeros((len(self.pulse_objs_list), self.max_bin_numbers))
        parameters_array = self.jnp.zeros((len(self.parameter_objs_list),))
        # Set the times
        for time_name in self.times_obj_dictionary:
            time_obj = self.times_obj_dictionary[time_name]
            if time_obj.is_optimization:
                time_obj.set_parameter(optimized_parameters_vector[time_obj.control_parameters_list])
        # Get the pulses and the timegrids
        for index, pulse in enumerate(self.pulse_objs_list):
            time_name = pulse.time_name
            pulse_array = pulse.get_pulse(
                optimized_parameters_vector[np.asarray(pulse.control_parameters_list)],
                final_time=self.times_obj_dictionary[time_name].get_time())
            pulses_array = pulses_array.at[index, :pulse.bins_number].set(self.jnp.asarray(pulse_array))
            time_grids_array = time_grids_array.at[index, :pulse.bins_number].set(self.jnp.asarray(pulse.time_grid))
        # Get the parameters
        for index, parameter in enumerate(self.parameter_objs_list):
            # parameters_array[index] = parameter.get_parameter(
            #     optimized_parameters_vector[parameter.control_parameters_list])
            parameters_array = parameters_array.at[index].set(self.jnp.asarray(parameter.get_parameter(
                optimized_parameters_vector[self.jnp.asarray(parameter.control_parameters_list)]), dtype=self.jnp.complex64))
        return pulses_array, time_grids_array, parameters_array

    def _get_controls_jax_obj(self, optimized_parameters_vector: np.array) -> np.array:
        """
        Sets the optimized control parameters and gets the controls in case of jax (AD) optimization

        :param np.array optimized_parameters_vector:
        :return: The pulses, time grids, and the parameters in three different lists of jax.numpy arrays.
        """
        pulses_array = self.jnp.zeros((len(self.pulse_objs_list), self.max_bin_numbers), dtype=self.jnp.complex64)
        time_grids_array = self.jnp.zeros((len(self.pulse_objs_list), self.max_bin_numbers), dtype=self.jnp.complex64)
        parameters_array = self.jnp.zeros((len(self.parameter_objs_list),), dtype=self.jnp.complex64)
        # Set the times
        for time_name in self.times_obj_dictionary:
            time_obj = self.times_obj_dictionary[time_name]
            if time_obj.is_optimization:
                time_obj.set_parameter(optimized_parameters_vector[time_obj.control_parameters_list])
        # Get the pulses and the timegrids
        for index, pulse in enumerate(self.pulse_objs_list):
            time_name = pulse.time_name
            pulse_array = pulse.get_pulse(
                optimized_parameters_vector[self.jnp.asarray(pulse.control_parameters_list)],
                final_time=self.times_obj_dictionary[time_name].get_time())
            pulses_array = pulses_array.at[index, :pulse.bins_number].set(self.jnp.asarray(pulse_array, dtype=self.jnp.complex64))
            time_grids_array = time_grids_array.at[index, :pulse.bins_number].set(self.jnp.asarray(pulse.time_grid, dtype=self.jnp.complex64))
        # Get the parameters
        for index, parameter in enumerate(self.parameter_objs_list):
            # parameters_array[index] = parameter.get_parameter(
            #     optimized_parameters_vector[parameter.control_parameters_list])
            parameters_array = parameters_array.at[index].set(self.jnp.asarray(parameter.get_parameter(
                optimized_parameters_vector[self.jnp.asarray(parameter.control_parameters_list)]), dtype=self.jnp.complex64))
        return pulses_array, time_grids_array, parameters_array

    def get_bare_controls_lists(self, optimized_parameters_vector: np.array) -> [list]:
        """
        Sets the optimized control parameters and gets the bare controls (without guess and shaping)

        :param np.array optimized_parameters_vector:
        :return: The bare OC pulses (without guess and shaping) in a list of numpy arrays.
        """
        bare_OC_pulses_list = []

        # Get the bare pulses and the timegrids
        for pulse in self.pulse_objs_list:
            time_name = pulse.time_name
            bare_OC_pulses_list.append(
                pulse.get_bare_pulse(optimized_parameters_vector[pulse.control_parameters_list],
                                     final_time=self.times_obj_dictionary[time_name].get_time()))

        return bare_OC_pulses_list



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import jax.scipy as jsp
import jax.numpy as jnp
from jax import jit

@jit
def fidelity_AD(sqrtm_a, b):
    # Calculate the eigenvalues
    # eig_vals = jnp.linalg.eigvals(jnp.matmul(jnp.matmul(sqrtm_a, b), sqrtm_a))
    # 
    eig_vals = jnp.linalg.eigvals(jnp.dot(jnp.dot(sqrtm_a, b), sqrtm_a))
    # Take the real part
    eig_vals = jnp.real(eig_vals)
    # Return the fidelity
    eig_vals = jnp.where(eig_vals > 0.0, eig_vals, 0.0)
    return jnp.power(jnp.sqrt(eig_vals).sum(), 2.0)

@jit
def fidelity_not_diff_due_to_sqrtm(rho, sigma):
    # Compute the square root of rho
    sqrt_rho = jsp.linalg.sqrtm(rho)

    # Compute the product sqrt_rho sigma sqrt_rho
    product = jnp.dot(jnp.dot(sqrt_rho, sigma), sqrt_rho)

    # Take the square root of the product
    sqrt_product = jsp.linalg.sqrtm(product)

    # Compute the trace of the square root of the product
    fid = jnp.trace(sqrt_product)

    # Take the real part to avoid numerical error leading to small imaginary component
    fid = jnp.real(fid)

    return fid


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np

from quocslib.stoppingcriteria.StoppingCriteria import StoppingCriteria


class NelderMeadStoppingCriteria(StoppingCriteria):
    """
    Class for the Nelder-Mead custom stopping criteria. Inherits from StoppingCriteria.
    """

    def __init__(self, stopping_criteria: dict):
        """
        Constructor of the Nelder-Mead stopping criteria class.

        :param dict stopping_criteria: Dictionary of specific stopping criteria
        """
        # Call to the super class constructor
        super().__init__(stopping_criteria)
        self.xatol = stopping_criteria.setdefault("xatol", 1e-14)
        self.fatol = stopping_criteria.setdefault("fatol", 1e-100)
        # self.is_converged = False

    def check_stopping_criteria(self,
                                sim: np.array = None,
                                fsim: np.array = None,
                                func_evaluations_single_direct_search: int = None) -> None:
        """
        Checks the stopping criteria for the Nelder-Mead algorithm.

        :param sim: Nelder-Mead simplex
        :param fsim: FoM values in the simplex
        :param func_evaluations_single_direct_search: Number of function evaluations in the direct search
        """
        if self.is_converged:
            return

        # self.is_converged, self.terminate_reason = self.check_func_eval_total(func_evaluations_single_direct_search)
        # if self.is_converged: return

        self.is_converged, self.terminate_reason = self.check_func_eval_single_direct_search(
            func_evaluations_single_direct_search)
        if self.is_converged:
            return

        self.is_converged, self.terminate_reason = self.check_simplex_criterion(sim)
        if self.is_converged:
            return

        self.is_converged, self.terminate_reason = self.check_f_size(fsim)
        if self.is_converged:
            return

        # self.is_converged, self.terminate_reason = self.check_goal_reached(fsim[0])
        # if self.is_converged: return

        # self.is_converged, self.terminate_reason = self.check_total_time_out()
        # if self.is_converged: return

        self.is_converged, terminate_reason = self.check_direct_search_time_out()
        if self.is_converged:
            self.terminate_reason = terminate_reason
            return



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np

from quocslib.pulses.superparameter.SuperParameterDistribution import (
    SuperParameterDistribution, )
from quocslib.tools.randomgenerator import RandomNumberGenerator


class Uniform(SuperParameterDistribution):
    def __init__(self,
                 basis_vectors_number,
                 super_parameter_distribution_dictionary,
                 rng: RandomNumberGenerator = None):
        """Spend here few words, compulsory arguments for the parent class"""
        self.basis_vectors_number = basis_vectors_number
        super().__init__(**super_parameter_distribution_dictionary)
        self.rng = rng

    def set_random_super_parameter(self):
        """Spend here few words here"""
        # Number of random super_parameters
        k = self.basis_vectors_number
        # Limits on the super_parameters
        a = self.lower_limit_w
        b = self.upper_limit_w
        # number of bins to be used in the calculation
        k_bins = (b - a) / k
        # generation of random super_parameters
        if self.rng is None:
            random_array = np.random.rand(k)
        else:
            random_array = self.rng.get_random_numbers(k)
        for i in range(k):
            self.w[i] = a + i * k_bins + k_bins * random_array[i]



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.utils.AbstractDump import AbstractDump


class DummyDump(AbstractDump):
    """Dummy class for dumping the controls"""

    def __init__(self, results_path: str = ".", date_time: str = ".", dump_format: str = "npz", **kwargs):
        pass

    def dump_controls(self, pulses: list = [], timegrids: list = [], parameters: list = [], **kwargs) -> None:
        """Do Nothing"""
        pass



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.utils.AbstractFoM import AbstractFoM
from quocslib.utils.inputoutput import writejsonfile
import numpy as np
import os
import time


class FilesUpdateFoM(AbstractFoM):
    """
    An evaluation method for the figure of merit via files exchange. The communication object accesses the
    get_FoM function.
    The get_FoM removes the "FoM.txt" file and creates a json, a txt or an npz file in the controls
    folder designed by the user.
    Finally, the read_FoM_values will wait and check for the figure of merit in the folder designed by the user.
    """
    def __init__(self,
                 controls_folder: str = ".",
                 is_split: bool = False,
                 file_extension: str = "json",
                 FoM_folder: str = ".",
                 max_time: float = 60,
                 **kwargs) -> None:
        """
        Constructor of the FilesUpdateFoM class

        :param str controls_folder: Path of the controls folder
        :param bool is_split: If True the controls are split in multiple files (not implemented yet)
        :param str file_extension: Extension of the controls file
        :param str FoM_folder: Path of the figure of merit folder
        :param int max_time: Maximum time in seconds to wait for the figure of merit evaluation
        :param kwargs: Other parameters
        """
        # Control folder
        self.controls_path = os.path.join(controls_folder, "controls.{0}".format(file_extension))
        # File extension
        self.file_extension = file_extension
        # Split the controls in multiple files
        # TODO Add this option in a future version
        self.is_split = is_split
        # FoM folder
        self.FoM_path = os.path.join(FoM_folder, "FoM.txt")
        # Maximum time in seconds to wait for the figure of merit evaluation
        self.max_time = max_time

    def get_FoM(self, pulses: list = [], timegrids: list = [], parameters: list = []) -> dict:
        """
        Write the controls in the controls.npz file, read the figure of merit in the FoM.txt file.

        :param list pulses: List of pulses
        :param list timegrids: List of timegrids
        :param list parameters: List of parameters
        :return dict: The figure of merit dictionary
        """
        print("Removing the previous {0} file if any".format(self.FoM_path))
        # If FoM.txt file exists remove it
        if os.path.exists(self.FoM_path):
            os.remove(self.FoM_path)
        # Write the pulses into a file or multiple files
        # TODO Multiple files extension
        print("Putting controls in {0}".format(self.controls_path))
        self.put_controls_into_user_path(pulses, timegrids, parameters)
        # Read the content of FoM.txt file
        print("Reading the {0} file ".format(self.FoM_path))
        return self.read_FoM_values()

    def read_FoM_values(self) -> dict:
        """
        Read the figure of merit FoM.txt file
        :return dict: The figure of merit dictionary
        """
        # Read the FoM
        FoM = None
        time_counter = 0
        time_counter_max = self.max_time
        # Check if figure of merit file exists
        while not os.path.exists(self.FoM_path):
            time_counter += 1
            time.sleep(0.5)
            if time_counter >= time_counter_max:
                return {"FoM": FoM, "status_code": -2}
        # Sleep to be sure the file is correctly close
        time.sleep(0.01)
        try:
            with open(self.FoM_path, "r") as FoM_file:
                FoM = float(str(FoM_file.readline()).strip())
        # TODO Add specific exception why the figure of merit is not readable
        except Exception as ex:
            print("Unhandled exception during FoM reading: {0}".format(ex.args))
            return {"FoM": FoM, "status_code": -3}
        return {"FoM": FoM}

    def put_controls_into_user_path(self, pulses_list: list, time_grids_list: list, parameters_list: list) -> None:
        """
        Save the controls in the controls.npz file.

        :param list pulses_list: List of np.arrays. One numpy array for each pulse
        :param list time_grids_list: List of np.arrays. One numpy array for each time grid
        :param list parameters_list: List of floats. One float for each parameter
        """
        # Choose between save as json file or txt file
        file_extension = self.file_extension
        if file_extension == "txt":
            self._put_controls_txt(pulses_list, time_grids_list, parameters_list)
        elif file_extension == "json":
            self._put_controls_json(pulses_list, time_grids_list, parameters_list)
        else:
            # TODO Return a negative status code
            print("The extension {0} is not recognized".format(file_extension))

    def _put_controls_txt(self, pulses_list: list, time_grids_list: list, parameters_list: list):
        """
        Orders the controls like pulse1, time_grid1, pulse2, time_grid2, ... para1, para2 ... and saves into a
        txt file.

        :param list pulses_list: List of np.arrays. One numpy array for each pulse
        :param list time_grids_list: List of np.arrays. One numpy array for each time grid
        :param list parameters_list: List of floats. One float for each parameter
        """
        with open(self.controls_path, "wb") as controls_file:
            for pulse, time_grid in zip(pulses_list, time_grids_list):
                np.savetxt(controls_file, [pulse], fmt="%f")
                np.savetxt(controls_file, [time_grid], fmt="%f")

            np.savetxt(controls_file, parameters_list, fmt="%f")

    def _put_controls_json(self, pulses_list: list, time_grids_list: list, parameters_list: list):
        """
        Saves the controls into a json file

        :param list pulses_list: List of np.arrays. One numpy array for each pulse
        :param list time_grids_list: List of np.arrays. One numpy array for each time grid
        :param list parameters_list: List of floats. One float for each parameter
        """
        # Save the pulses and the respective timegrids into a dictionary
        controls_dict = {}
        pulse_index = 1
        # The pulses are saved like pulse#
        # The timegrids are saved like time_grids#
        for pulse, time_grid in zip(pulses_list, time_grids_list):
            controls_dict["pulse" + str(pulse_index)] = pulse
            controls_dict["time_grid" + str(pulse_index)] = time_grid
            pulse_index += 1
        parameter_index = 1
        # The parameters are saved like parameter#
        for parameter in parameters_list:
            controls_dict["parameter" + str(parameter_index)] = parameter
            parameter_index += 1
        writejsonfile(self.controls_path, controls_dict)



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from scipy.special import erf

from quocslib.pulses.BasePulse import BasePulse
from quocslib.pulses.basis.ChoppedBasis import ChoppedBasis
from quocslib.tools.randomgenerator import RandomNumberGenerator


class SigmoidSigmas(ChoppedBasis):
    amplitude_variation: float
    optimized_control_parameters: np.ndarray
    optimized_super_parameters: np.ndarray
    time_grid: np.ndarray

    def __init__(self, map_index: int, pulse_dictionary: dict, rng: RandomNumberGenerator = None, is_AD: bool = False):
        """
        Constructor of the Sigmoid basis class with adjustable sigmas. It calls the constructor of the parent
        class ChoppedBasis.

        :param int map_index: Index number to use to get the control parameter.
        :param dict pulse_dictionary: The dictionary of the pulse is defined here. Only the basis dictionary is used.
        """
        basis_dict = pulse_dictionary["basis"]
        # Frequencies number i.e. the basis vector number in the pulse parametrization
        self.super_parameter_number = basis_dict.setdefault("basis_vector_number", 1)

        # define offset from side and sigma
        self.sigma = basis_dict.setdefault("sigma", 10**(-10))
        self.offset = basis_dict.setdefault("offset", -10*self.sigma)
        # sigma initial variation
        self.sigma_var = basis_dict.setdefault("sigma_variation", 1.0)

        # Number of control parameters to be optimized
        self.control_parameters_number = 2 * self.super_parameter_number + 1

        # Constructor of the parent class Chopped BasisX
        super().__init__(map_index=map_index, rng=rng, is_AD=is_AD, **pulse_dictionary)

        # Define scale and offset coefficients
        self.scale_coefficients = self.amplitude_variation * np.ones((self.control_parameters_number, ))
        self.offset_coefficients = np.zeros((self.control_parameters_number, ))

    def _get_shaped_pulse(self) -> np.array:
        """
        Definition of the pulse parametrization. It is called at every function evaluation to build the pulse and
        return it as an array.

        :return np.array: The pulse as an array.
        """

        # initilize pulse
        pulse = np.zeros(self.bins_number)
        # get final time
        final_time = self.final_time

        # amplitudes
        Aopti = self.optimized_control_parameters[0:int((len(self.optimized_control_parameters)+1)/2)]
        # sigma variation
        del_sig = abs(self.optimized_control_parameters[int((len(self.optimized_control_parameters)+1)/2):]) * self.sigma_var
        # times
        taus = self.super_parameter_distribution_obj.w

        sigmas = del_sig + self.sigma
        # adjusting the sigmas
        if self.offset >= 10**(-10):
            # making sure it stays in range if the offset is unequal zero
            kappa = self.sigma/self.offset
            for i in range(len(sigmas)):
                del_sigma_max = (final_time/2 - abs(final_time/2-taus[i])) * kappa - self.sigma
                sigmas[i] = self.sigma + del_sigma_max*(1-np.cos(np.pi*del_sig[i]))/2

        # optimize pulse
        t = self.time_grid
        for ii in range(self.super_parameter_number):
            pulse += (Aopti[ii + 1] / 2 * (erf((t - taus[ii]) / (np.sqrt(2) * sigmas[ii])) + 1))
        pulse += Aopti[0] / 2 * (erf((t - self.offset) / (np.sqrt(2) * self.sigma)) + 1)
        pulse += (-np.sum(Aopti) / 2 * (erf((t - (final_time - self.offset)) / (np.sqrt(2) * self.sigma)) + 1))
        return pulse


import jax.numpy as jnp


def fidelity_funct(rho_evolved, rho_aim):
    """
    Fidelity function via the Hilbert-Schmidt inner product
    :param rho_evolved:
    :param rho_aim:
    :return float: State overlap
    """
    return jnp.abs(jnp.trace(rho_evolved.conj().T @ rho_aim))



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np

from quocslib.tools.randomgenerator import RandomNumberGenerator

try:
    import jax
    import jax.numpy as jnp
    import jax.scipy.optimize as sopt
    jax.config.update('jax_enable_x64', True)
except:
    raise ImportError

from scipy import optimize

from quocslib.Controls import Controls
from quocslib.optimizationalgorithms.OptimizationAlgorithm import OptimizationAlgorithm


class ADAlgorithm(OptimizationAlgorithm):
    """
    This is an implementation of the automatic differentiation (AD) algorithm for open-loop optimal control.
    The important functions are:
    * the constructor with the optimization dictionary and the communication object as parameters
    * run : The main loop for optimal control
    * _get_response_for_client : return info about the goodness of the controls and errors if any
    * _get_controls : return the set of controls as a dictionary with pulses, parameters, and times as keys
    * _get_final_results: return the final result of the optimization algorithm
    """

    def __init__(self, optimization_dict: dict = None, communication_obj=None, FoM_object=None, **kwargs):
        """
        This is the implementation of the AD algorithm. All the arguments in the constructor are passed to the
        OptimizationAlgorithm constructor except for the FoM_object, which is used to compute the figure of merit.

        :param optimization_dict: dictionary with the settings for the optimization
        :param communication_obj: object to communicate with the client
        :param FoM_object: object to compute the figure of merit
        """
        super().__init__(communication_obj=communication_obj, optimization_dict=optimization_dict)
        ###########################################################################################
        # Optimal algorithm variables if any
        ###########################################################################################
        self.FoM_object = FoM_object
        self.FoM_list = []
        self.sys_type = optimization_dict.setdefault("sys_type", "StateTransfer")
        # set stopping criteria
        self.max_num_si = int(optimization_dict["algorithm_settings"].setdefault("super_iteration_number", 1))
        self.stopping_crit = optimization_dict["algorithm_settings"].setdefault("stopping_criteria", {})
        self.max_fun_evals = self.stopping_crit.setdefault("max_eval_total", 10 ** 10)
        self.ftol = self.stopping_crit.setdefault("ftol", 1e-6)
        self.gtol = self.stopping_crit.setdefault("gtol", 1e-6)
        self.maxls = self.stopping_crit.setdefault("maxls", 20)  # 20 is default acc. to documentation of scipy

        self.is_record = False

        alg_parameters = optimization_dict["algorithm_settings"]
        # Seed for the random number generator
        if "random_number_generator" in alg_parameters:
            try:
                seed_number = alg_parameters["random_number_generator"]["seed_number"]
                self.rng = RandomNumberGenerator(seed_number=seed_number)
            except (TypeError, KeyError):
                default_seed_number = np.random.randint(0, 10000)
                message = "Seed number must be an integer value. Set {0} as a seed numer for this optimization".format(
                    default_seed_number)
                self.rng = RandomNumberGenerator(seed_number=default_seed_number)
                self.comm_obj.print_logger(message, level=30)
        else:
            default_seed_number = np.random.randint(0, 10000)
            message = "Seed number must be an integer value. Set {0} as a seed numer for this optimization".format(
                default_seed_number)
            self.rng = RandomNumberGenerator(seed_number=default_seed_number)
            self.comm_obj.print_logger(message, level=30)

        ###########################################################################################
        # Pulses, Parameters, Times object
        ###########################################################################################
        # Define array objects for the gradient calculation
        self.controls = Controls(
            optimization_dict["pulses"],
            optimization_dict["times"],
            optimization_dict["parameters"],
            rng=self.rng,
            is_AD=True,
        )

        ###########################################################################################
        # Other useful variables
        ###########################################################################################
        self.FoM_list: list = []
        self.iteration_number_list: list = []
    
    def inner_routine_call(self, optimized_control_parameters: jnp.array):
        """
        Wrapper function for the _routine_call to make it consistent with the other optimization algorithms.

        :param optimized_control_parameters: array with the optimized control parameters
        :return float: FoM
        """
        self.is_record = False
        FoM = self._routine_call(optimized_control_parameters=optimized_control_parameters, iterations=0)
        return FoM

    def value_grad(self):
        """ Return value and grad with jax """
        jax_function = jax.value_and_grad(self.inner_routine_call)
        return jax_function

    def get_gradient(self, optimized_control_parameters: np.array):
        """ Used to calculate the gradient from the FoM object's get_FoM function """
        [pulses, timegrids, parameters] = self.controls.get_controls_lists(optimized_control_parameters)
        return self.FoM_factor * self.FoM_object.get_FoM(pulses=pulses,
                                                         parameters=parameters,
                                                         timegrids=timegrids)["FoM"]

    def run(self):
        """ Main loop for the optimization algorithm. It runs the super iterations and the inner iterations. """
        for super_it in range(1, self.max_num_si + 1):
            # Set super iteration number
            self.super_it = super_it
            # Initialize the random super_parameters
            self.controls.select_basis()
            #
            self.inner_run()
            #
            self.controls.update_base_controls(self.best_xx)

    def inner_run(self):
        """ Inner routine call that runs the optimization algorithm. """
        heuristic_coeff = 1.0
        random_variation = heuristic_coeff * 2 * (0.5 - self.rng.get_random_numbers(self.controls.get_control_parameters_number()))
        # Scale it according to the amplitude variation
        initial_variation = random_variation * self.controls.get_sigma_variation()
        # Define the initial
        init_xx = self.controls.get_mean_value() + initial_variation

        def get_gradient(x):
            return np.array(jax.jit(jax.grad(self.get_gradient))(x))

        def f_call(x):
            return np.array(self.inner_routine_call(x))

        results = optimize.minimize(f_call,
                                    x0=init_xx,
                                    jac=get_gradient,
                                    method='L-BFGS-B',  # method='BFGS', # method='L-BFGS-B',
                                    options={
                                        'disp': True,
                                        'ftol': self.ftol,
                                        'maxfun': self.max_fun_evals,
                                        'gtol': self.gtol,
                                        'maxls': self.maxls
                                    })

        # Print L-BFGS-B results in the log file
        self.comm_obj.print_logger(results, level=20)
        # Update the controls with the best ones found so far
        # self.controls.update_base_controls(results.x)
        # self.controls.update_base_controls(self.best_xx)

    def _get_controls(self, optimized_control_parameters: jnp.array) -> dict:
        """
        Get the controls dictionary from the optimized control parameters

        :param jnp.array optimized_control_parameters: the array of optimize control parameters
        :return dict: returns a dict that contains the pulses, parameters and timegrids
        """
        # jax.debug.print("_get_controls, optimized_control_parameters: {}", optimized_control_parameters)
        [pulses, timegrids, parameters] = self.controls.get_controls_lists(optimized_control_parameters)
        #
        controls_dict = {
            "pulses": pulses,
            "parameters": parameters,
            "timegrids": timegrids,
        }
        return controls_dict

    def _get_final_results(self) -> dict:
        """ Return a dictionary with final results to put into a dictionary """
        final_dict = {
            "Figure of merit": self.FoM_factor * self.best_FoM,
            "total number of function evaluations": self.iteration_number,
        }
        return final_dict

    def _get_response_for_client(self) -> dict:
        """ Return a dictionary with the response for the client """
        FoM = self.FoM_dict["FoM"]
        status_code = self.FoM_dict.setdefault("status_code", 0)
        if self.get_is_record(FoM):
            message = "New record achieved. Previous FoM: {FoM}, new best FoM : {best_FoM}".format(FoM=self.FoM_factor*self.best_FoM,
                                                                                                   best_FoM=self.FoM_factor*FoM)
            self.comm_obj.print_logger(message=message, level=20)
            self.best_FoM = float(FoM)
            self.best_xx = self.xx.copy()
            self.is_record = True
        response_dict = {
            "is_record": self.is_record,
            "FoM": FoM,
            "iteration_number": self.iteration_number,
            "status_code": status_code
        }
        # Load the current figure of merit and iteration number in the summary list
        if status_code == 0:
            self.FoM_list.append(self.FoM_factor*FoM)
            self.iteration_number_list.append(self.iteration_number)
        return response_dict



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np
from scipy.optimize import minimize
import scipy

from quocslib.Controls import Controls
from quocslib.optimizationalgorithms.OptimizationAlgorithm import OptimizationAlgorithm

from quocslib.tools.linearalgebra import commutator
from quocslib.tools.randomgenerator import RandomNumberGenerator


class GRAPEAlgorithm(OptimizationAlgorithm):
    """
    This is an implementation of the gradient ascent pulse engineering (GRAPE) algorithm for open-loop optimal control.
    The important functions are:
    * the constructor with the optimization dictionary and the communication object as parameters
    * run : The main loop for optimal control
    * _get_response_for_client : return info about the goodness of the controls and errors if any
    * _get_controls : return the set of controls as a dictionary with pulses, parameters, and times as keys
    * _get_final_results: return the final result of the optimization algorithm
    """
    def __init__(self, optimization_dict: dict = None, communication_obj=None, FoM_object=None, **kwargs):
        """
        This is the implementation of the GRAPE algorithm. All the arguments in the constructor are passed to the
        OptimizationAlgorithm constructor except for the FoM_object, which is used to compute the figure of merit.

        :param optimization_dict: dictionary with the settings for the optimization
        :param communication_obj: object to communicate with the client
        :param FoM_object: object to compute the figure of merit
        """
        super().__init__(communication_obj=communication_obj, optimization_dict=optimization_dict)
        ###########################################################################################
        # Optimal algorithm variables
        ###########################################################################################
        # Get functions and variable for the gradient optimization
        self.propagator_func = FoM_object.get_propagator
        self.initial_state = FoM_object.get_initial_state()
        self.target_state = FoM_object.get_target_state()
        self.drift_Hamiltonian = FoM_object.get_drift_Hamiltonian()
        self.control_Hamiltonians = FoM_object.get_control_Hamiltonians()
        self.FoM_list = []
        self.sys_type = optimization_dict.setdefault("sys_type", "StateTransfer")

        self.is_record = False

        # set stopping criteria
        self.stopping_crit = optimization_dict["algorithm_settings"].setdefault("stopping_criteria", {})
        self.max_fun_evals = self.stopping_crit.setdefault("max_eval_total", 10**10)
        self.ftol = self.stopping_crit.setdefault("ftol", 1e-6)
        self.gtol = self.stopping_crit.setdefault("gtol", 1e-6)
        self.maxls = self.stopping_crit.setdefault("maxls", 20)  # 20 is default acc. to documentation of scipy

        alg_parameters = optimization_dict["algorithm_settings"]
        # Seed for the random number generator
        seed_number = 2022
        if "random_number_generator" in alg_parameters:
            try:
                seed_number = alg_parameters["random_number_generator"]["seed_number"]
            except (TypeError, KeyError):
                seed_number = 2022
                message = "Seed number must be an integer value. Set {0} as a seed numer for this optimization".format(
                    seed_number)
                self.comm_obj.print_logger(message, level=30)
        self.rng = RandomNumberGenerator(seed_number=seed_number)

        ###########################################################################################
        # Pulses, Parameters, Times object
        ###########################################################################################
        # Define array objects for the gradient calculation
        self.controls = Controls(
            optimization_dict["pulses"],
            optimization_dict["times"],
            optimization_dict["parameters"],
            rng=self.rng,
        )
        ###########################################################################################
        # Objects for gradient optimization
        ###########################################################################################
        # Get the bins number of the first pulse
        self.n_slices = n_slices = self.controls.pulse_objs_list[0].bins_number
        self.propagator_storage = np.array([np.zeros_like(self.drift_Hamiltonian) for _ in range(n_slices)])
        self.rho_storage = np.array([np.zeros_like(self.target_state) for _ in range(n_slices + 1)])
        self.rho_storage[0] = self.initial_state
        self.corho_storage = np.array([np.zeros_like(self.target_state) for _ in range(n_slices + 1)])
        self.corho_storage[-1] = self.target_state
        ###########################################################################################
        # Other useful variables
        ###########################################################################################
        self.FoM_list: list = []
        self.iteration_number_list: list = []

    def _get_response_for_client(self) -> dict:
        """
        This function returns the response for the client.

        :return dict: Information dictionary with the FoM, iteration number, status code, and if the FoM is a new
        record
        """
        FoM = self.FoM_dict["FoM"]
        status_code = self.FoM_dict.setdefault("status_code", 0)
        if self.get_is_record(FoM):
            message = "New record achieved. Previous FoM: {FoM}, new best FoM : {best_FoM}".format(FoM=self.FoM_factor*self.best_FoM,
                                                                                                   best_FoM=self.FoM_factor*FoM)
            self.comm_obj.print_logger(message=message, level=20)
            self.best_FoM = FoM
            self.best_xx = self.xx.copy()
            self.is_record = True
        response_dict = {
            "is_record": self.is_record,
            "FoM": FoM,
            "iteration_number": self.iteration_number,
            "status_code": status_code
        }
        # Load the current figure of merit and iteration number in the summary list of dCRAB
        if status_code == 0:
            self.FoM_list.append(self.FoM_factor*FoM)
            self.iteration_number_list.append(self.iteration_number)
        return response_dict

    def get_gradient(self, optimized_control_parameters: np.array):
        """
        Get the gradient from the propagators calculated in the FoM object

        :param optimized_control_parameters: optimized control parameters
        :return: gradients
        """
        # Calculate the controls
        [pulses, timegrids, parameters] = self.controls.get_controls_lists(optimized_control_parameters)
        # Pass the controls to the get propagator function
        U_store = self.propagator_func(pulses_list=pulses, time_grids_list=timegrids, parameters_list=parameters)
        n_slices = self.n_slices
        sys_type = self.sys_type
        rho_store = self.rho_storage
        corho_store = self.corho_storage

        time_grid = timegrids[0]
        # dt = time_grid[1] - time_grid[0]
        dt = time_grid[-1] / len(time_grid)

        # Number of control Hamiltonians
        K = len([self.control_Hamiltonians])
        # control hamiltonians
        B = [self.control_Hamiltonians]
        # Forward and backward propagation
        for t in range(n_slices):
            U = U_store[t]
            # depending on system type do different evolution
            if sys_type == "StateTransfer":
                rho_store[t + 1] = U @ rho_store[t] @ U.T.conj()
            else:
                rho_store[t + 1] = U @ rho_store[t]
        for t in reversed(range(n_slices)):
            U = U_store[t]
            if sys_type == "StateTransfer":
                corho_store[t] = U.T.conj() @ corho_store[t + 1] @ U
            else:
                corho_store[t] = corho_store[t + 1] @ U.T.conj()
        # Calculate the gradient
        grads = np.zeros((K, n_slices))
        for k in range(K):
            for t in range(n_slices):
                if sys_type == "StateTransfer":
                    g = (1j * dt * corho_store[t].T.conj() @ commutator(B[k], rho_store[t]))
                    grads[k, t] = np.real(np.trace(g))
                else:
                    grads[k, t] = 0.0
        grads = grads.flatten()
        # Return the gradient to the main updating function
        return grads

    def inner_routine_call(self, optimized_control_parameters: np.array):
        """Function evaluation call for the L-BFGS-B algorithm"""
        self.is_record = False
        grads = self.get_gradient(optimized_control_parameters=optimized_control_parameters)
        FoM = self._routine_call(optimized_control_parameters=optimized_control_parameters, iterations=0)
        return FoM, grads

    def run(self) -> None:
        """Starts the main loop of the optimization"""
        # Initial set of random parameters
        # I have to use random parameters to avoid initial local trap (null gradient)
        # Create array with values between [-1.0, 1.0]
        random_variation = 2 * (0.5 - self.rng.get_random_numbers(self.controls.get_control_parameters_number()))
        # Scale it accordingly to the amplitude variation
        initial_variation = random_variation * self.controls.get_sigma_variation()
        # Define the initial
        init_xx = self.controls.get_mean_value() + initial_variation
        # Optimization with L-BFGS-B
        results = scipy.optimize.minimize(self.inner_routine_call,
                                          init_xx,
                                          method="L-BFGS-B",
                                          jac=True,
                                          options={
                                              'ftol': self.ftol,
                                              'maxfun': self.max_fun_evals,
                                              'gtol': self.gtol,
                                              'maxls': self.maxls
                                          })
        # Print L-BFGS-B results in the log file
        self.comm_obj.print_logger(results, level=20)
        # Update the controls with the best ones found so far
        self.controls.update_base_controls(self.best_xx)

    def _get_controls(self, xx: np.array) -> dict:
        """
        Get the controls dictionary from the optimized control parameters

        :param xx: optimized control parameters
        :return dict : dictionary with the controls
        """
        [pulses, timegrids, parameters] = self.controls.get_controls_lists(xx)

        controls_dict = {
            "pulses": pulses,
            "parameters": parameters,
            "timegrids": timegrids,
        }
        return controls_dict

    def _get_final_results(self) -> dict:
        """
        Return a dictionary with final results

        :return dict: dictionary with final results
        """
        final_dict = {
            "Figure of merit": self.FoM_factor * self.best_FoM,
            "nfev": self.iteration_number,
        }
        return final_dict



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from scipy.linalg import expm, norm
from quocslib.utils.AbstractFoM import AbstractFoM


def hamiltonian_d1_d2_2fields(amplitude_t, phase_t, delta1=0.0, delta2=0.0):
    """
    The Hamiltonian to use for the OneQubit problem with 2 driving fields.
    :param amplitude_t: amplitude of the driving field
    :param phase_t: phase between sigma_x and sigma_y
    :param delta1: detuning on the energy levels
    :param delta2: detuning on the drive
    :return: The Hamiltonian
    """
    sigma_x = np.array([[0, 1], [1, 0]], dtype="complex")
    sigma_y = np.array([[0, -1j], [1j, 0]], dtype="complex")
    sigma_z = np.array([[1, 0], [0, -1]], dtype="complex")

    ham_t = delta1 * sigma_z / 2 + amplitude_t * (1 + delta2) * (np.cos(phase_t) * sigma_x + np.sin(phase_t) * sigma_y)
    return ham_t


class OneQubit2Fields(AbstractFoM):
    """
    This class implements the one qubit problem with two time-dependent function in the drive
    as an example FoM class.
    """
    def __init__(self, args_dict: dict = None):
        if args_dict is None:
            args_dict = {}

        self.psi_target = np.asarray(
            eval(args_dict.setdefault("target_state", "[1.0/np.sqrt(2), -1j/np.sqrt(2)]")),
            dtype="complex",
        )
        self.psi_0 = np.asarray(eval(args_dict.setdefault("initial_state", "[1.0, 0.0]")), dtype="complex")

        # two constant detuning values to use in the Hamiltonian
        self.delta1 = args_dict.setdefault("delta1", 0.1)
        self.delta2 = args_dict.setdefault("delta2", 0.1)

        # Noise in the figure of merit
        self.is_noisy = args_dict.setdefault("is_noisy", False)
        self.noise_factor = args_dict.setdefault("noise_factor", 0.05)

    def get_FoM(self, pulses: list = [], parameters: list = [], timegrids: list = []) -> dict:
        """
        This function calculates the figure of merit for the one qubit problem with two fields.
        :param pulses:
        :param parameters:
        :param timegrids:
        :return dict: The FoM and the standard deviation in a dictionary
        """
        amplitude = np.asarray(pulses[0])
        phase = np.asarray(pulses[1])
        timegrid = np.asarray(timegrids[0])
        dt = timegrid[1] - timegrid[0]
        U = self._time_evolution(amplitude, phase, dt, self.delta1, self.delta2)
        psi_f = np.matmul(U, self.psi_0)
        infidelity = 1.0 - self._get_fidelity(self.psi_target, psi_f)

        std = 0.0
        if self.is_noisy:
            noise = (self.noise_factor * 2 * (0.5 - np.random.rand(1, )[0]))
            infidelity += noise
            std = self.noise_factor * 0.6827  # one std contains 68.28% of values

        return {"FoM": np.abs(infidelity), "std": std}

    @staticmethod
    def _time_evolution(amplitude, phase, dt, delta1=0.0, delta2=0.0):
        """
        This function calculates the time evolution operator for the one qubit problem with two fields.
        :param amplitude:
        :param phase:
        :param dt:
        :param delta1:
        :param delta2:
        :return:
        """
        U = np.identity(2)
        for ii in range(amplitude.size):
            ham_t = hamiltonian_d1_d2_2fields(amplitude[ii],
                                              phase[ii],
                                              delta1,
                                              delta2)
            U_temp = U
            U = np.matmul(expm(-1j * ham_t * dt), U_temp)
        return U

    @staticmethod
    def _get_fidelity(psi1, psi2):
        """
        This function calculates the fidelity between two states.
        :param psi1:
        :param psi2:
        :return float: Fidelity between psi1 and psi2
        """
        return np.abs(np.dot(psi1.conj().T, psi2))**2 / (norm(psi1) * norm(psi2))



import numpy as np


class OrnsteinUhlenbeck:
    """
    Capture the important features of OU noise based on the correlation time (tau) and the noise strength (c) and the
    start value (x0).
    """
    def __init__(self, tau=1.0, c=1.0, x0=0.0, dt=0.01):
        self.tau = tau
        self.c = c
        self.x0 = x0
        self.dt = dt

    def get_noise_realization(self, T):
        """
        Get a single noise realization
        :param T: total time of the noise realization
        :return: a single noise realization
        """
        n_slices = int(np.floor(T / self.dt))
        output = np.zeros(n_slices)
        output[0] = self.x0
        mu = np.exp(-(1 / self.tau) * self.dt)
        sigma = self.c * self.tau / 2 * (1 - np.exp(-2 / self.tau * self.dt))

        for i in range(n_slices - 1):
            output[i + 1] = output[i] * mu + np.sqrt(sigma) * np.random.randn()
        return output

    def get_ensemble_noise_realization(self, T, n_traj, **kwargs):
        """
        Get an ensemble of noise realizations
        :param T: total time of the noise realizations
        :param n_traj: number of noise trajectories
        :param kwargs:
        :return: an ensemble of noise realizations
        """
        n_slices = int(np.floor(T / self.dt))

        # now if x0 is a function then we sample from it
        # if x0 is an array we just use it
        # if callable(x0_distribution):
        #     x0_samples = np.array([x0_distribution(**kwargs) for t in range(n_traj)])
        # else:
        #     x0_samples = x0_distribution

        # create the output array
        output = np.zeros((n_traj, n_slices))
        for i in range(n_traj):
            output[i, :] = self.get_noise_realization(T)

        return output



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import os
import time
from typing import List

from quocslib.utils.AbstractFoM import AbstractFoM
from quocslib.utils.DummyDump import DummyDump
from quocslib.handleexit.AbstractHandleExit import AbstractHandleExit
from quocslib.tools.logger import create_logger
from quocslib.utils.inputoutput import writejsonfile
from quocslib import __VERSION__ as quocslib_version


class AllInOneCommunication:
    def __init__(self,
                 interface_job_name: str = "OptimizationTest",
                 FoM_obj: AbstractFoM = None,
                 handle_exit_obj: AbstractHandleExit = None,
                 dump_attribute: callable = DummyDump,
                 comm_signals_list: List[List] = None,
                 create_logfile: bool = True,
                 console_info: bool = True,
                 dump_format: str = "npz",
                 optimization_direction: str = "minimization"):
        """
        In case the user chooses to run the optimization in his device, this class is used by the OptimizationAlgorithm.
        The objects to dump the results, calculate the figure of merit, and the logger are created here. 

        :param str interface_job_name: Name decided by the Client. It is change in the constructor adding the current
        time to ensure univocity
        :param AbstractFoM FoM_obj: object for the figure of merit evaluation. Have a look to the abstract class for
        more info
        :param AbstractHandleExit handle_exit_obj: Collect any error during the optimization and check when the
        communication is finished to communicate with the client interface
        :param [list, list, list] comm_signals_list: List containing the signals to the gui
        :param create_logfile: Boolean to create the log file or not
        :param console_info: Boolean to show info in the console or not
        """
        # Communication signals
        if comm_signals_list is None:
            self.message_signal, self.FoM_plot_signal, self.controls_update_signal = (None, None, None)
        else:
            (self.message_signal, self.FoM_plot_signal, self.controls_update_signal) = comm_signals_list
        # Pre job name
        pre_job_name = interface_job_name
        # Datetime for 1-1 association
        self.date_time = str(time.strftime("%Y%m%d_%H%M%S"))
        # Client job name to send to the Server
        self.client_job_name = self.date_time + "_" + pre_job_name
        ###
        # Logging, Results, Figure of merit evaluation ...
        ###
        # Optimization folder
        optimization_folder = "QuOCS_Results"
        self.results_path = os.path.join(os.getcwd(), optimization_folder, self.client_job_name)
        if not os.path.isdir(os.path.join(os.getcwd(), optimization_folder)):
            os.makedirs(os.path.join(os.getcwd(), optimization_folder))
        # Create the folder for logging and results
        os.makedirs(self.results_path)
        # Write the current quocs lib version in the file
        with open(os.path.join(self.results_path, "quocs_version.txt"), "w") as version_file:
            version_file.write("QuOCS library version: {0}".format(quocslib_version))
        # Create logging object
        self.logger = create_logger(self.results_path, self.date_time, create_logfile=create_logfile, console_info=console_info)
        # Print function evaluation and figure of merit
        self.print_general_log = True
        # Figure of merit object
        self.FoM_obj = FoM_obj
        # TODO Thinks whether it is a good idea dumping the results
        # Dumping data object
        self.dump_obj = dump_attribute(self.results_path, self.date_time, dump_format)
        # Handle exit object
        self.he_obj = handle_exit_obj
        # Initialize the control dictionary
        self.controls_dict = {}
        # Initialize the controls names dictionary
        self.controls_names_dict = {}
        # multiply by -1 if the optimization is a maximization
        self.FoM_factor = 1
        if optimization_direction == "maximization":
            self.FoM_factor = -1

    def print_logger(self, message: str = "", level: int = 20):
        """Print a message in the log"""
        if level <= 10:
            self.logger.debug(message)
        elif 10 < level <= 20:
            self.logger.info(message)
        elif 20 < level <= 30:
            self.logger.warning(message)
        else:
            self.logger.error(message)

    def send_message(self, message):
        """Send a message to the interface"""
        if self.message_signal is not None:
            self.message_signal.emit(message)

    def print_optimization_dictionary(self, optimization_dictionary: dict) -> None:
        """Print optimization dictionary into a file"""
        writejsonfile(os.path.join(self.results_path, self.date_time + "_" + "optimization_dictionary.json"),
                      optimization_dictionary)

    def get_user_running(self) -> bool:
        """Check if the user stopped the optimization"""
        self.logger.debug("User running: {0}".format(self.he_obj.is_user_running))
        return self.he_obj.is_user_running

    def set_is_running_state(self, value: bool) -> None:
        """Set if the optimization is running"""
        self.he_obj.set_is_user_running(value)
        self.logger.info("Setting is_running state: {0}".format(self.he_obj.is_user_running))

    def send_controls(self, controls_dict: dict) -> None:
        """Set the controls for FoM calculation and notify the gui"""
        self.controls_dict = controls_dict
        if self.controls_update_signal is not None:
            self.controls_update_signal.emit(controls_dict["pulses"], controls_dict["timegrids"],
                                             controls_dict["parameters"])

    def set_controls_names(self, controls_names_dict: dict) -> None:
        """Set the controls names dictionary"""
        self.controls_names_dict = controls_names_dict

    def get_data(self) -> dict:
        """
        Calculate the figure of merit and return a dictionary with all the arguments

        :return dict: {"FoM_values": {"FoM": float, ...}}
        """
        self.logger.debug("User running: {0}".format(self.he_obj.is_user_running))
        FoM_dict = self.FoM_obj.get_FoM(**self.controls_dict)
        # multiply by -1 if the optimization is a maximization
        FoM_dict["FoM"] *= self.FoM_factor
        # set the status of the FoM to 0 if it does not exist already
        status_code = FoM_dict.setdefault("status_code", 0)
        # if the user passes a different status code than zero, stop the optimization
        if status_code != 0:
            self.he_obj.is_user_running = False
        return {"FoM_values": FoM_dict}

    def send_FoM_response(self, response_for_client: dict) -> None:
        """
        Emit signal to the Client Interface and dump the results in case any

        :param dict response_for_client: It is a dictionary defined in the optimal algorithm
        :return:
        """
        iteration_number, FoM = (response_for_client["iteration_number"], response_for_client["FoM"])
        status_code = response_for_client.setdefault("status_code", 0)
        # Check for interrupting signals
        if status_code != 0:
            self.logger.info("The optimization was interrupted with status code: {0}"
                             " at iteration {1}".format(status_code, iteration_number))
            # Set the user running to False in order to not continue with the next iteration
            self.he_obj.is_user_running = False
            return
        self._print_general_log(iteration_number, FoM)
        self.update_controls(**response_for_client)
        if self.FoM_plot_signal is not None:
            self.FoM_plot_signal.emit(iteration_number, FoM)

    def _print_general_log(self, iteration_number: int, FoM: float):
        """Print the general log at each function evaluation"""
        if self.print_general_log:
            self.logger.info("Function evaluation number: {0}, FoM: {1}".format(iteration_number, self.FoM_factor*FoM))

    def update_controls(self, **response_for_client) -> None:
        """External call to update the controls"""
        if response_for_client is None:
            response_for_client = {}
        self.dump_obj.dump_controls(**self.controls_dict, **self.controls_names_dict, **response_for_client)

    def end_communication(self, results_dict: dict) -> None:
        """
        Report the final results

        :param dict results_dict: It is a dictionary defined in the optimal algorithm with all the data to display at
        the end of the optimization process
        :return:
        """
        # Print final results data
        for el in results_dict:
            self.logger.info("{0} : {1}".format(el, results_dict[el]))

    def assign_job(self) -> None:
        """Do nothing"""
        pass

    def send_communication_data(self) -> None:
        """Do nothing"""
        pass

    def update_init_msg_server(self, upd_name=None) -> None:
        """Do nothing"""
        pass

    def check_msg_client(self) -> None:
        """Do nothing"""
        pass

    def update_msg_server(self) -> None:
        """Do nothing"""
        pass



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import jax.numpy as jnp
import jax

def create_mask(trace_over_positions, num_subsystems=5):
    """
    Create a mask for the transpose operation, given the positions of subsystems to trace over in a system of 5 subsystems.
    """
    if any(pos < 0 or pos >= num_subsystems for pos in trace_over_positions):
        raise ValueError('Positions must be within the range of the number of subsystems')

    # Create a permutation of subsystems excluding the ones to be traced over
    remaining_positions = [i for i in range(num_subsystems) if i not in trace_over_positions]

    # Initialize the mask with the remaining positions
    mask = remaining_positions + [pos + num_subsystems for pos in remaining_positions]

    # Add the positions of the subsystems to be traced over at the end
    for pos in trace_over_positions:
        mask.append(pos)
        mask.append(pos + num_subsystems)

    return mask

def jax_partial_trace_transform(subsystem_dims, trace_over_positions):
    """
    Compute the partial trace of a quantum state over specified subsystems using JAX.
    """
    num_subsystems = len(subsystem_dims)
    # mask = create_mask(trace_out_positions, num_subsystems)
    mask = create_mask(trace_over_positions, num_subsystems)

    # Final reshape
    remaining_dims = jnp.array([subsystem_dims[i] for i in range(num_subsystems) if i not in trace_over_positions])
    remaining_dim = jnp.prod(remaining_dims)
    @jax.jit
    def partial_trace(state: jnp.ndarray):
        """ Return the partial trace of the state over the specified subsystems. """
        state_reshaped = jnp.reshape(state, subsystem_dims * 2)
        reordered_state = jnp.transpose(state_reshaped, mask)
        for _ in range(len(trace_over_positions)):
            reordered_state = jnp.trace(reordered_state, axis1=-2, axis2=-1)
        return jnp.reshape(reordered_state, [remaining_dim, remaining_dim])

    return partial_trace



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
import jax.numpy as jnp
from quocslib.utils.AbstractFoM import AbstractFoM
from quocslib.timeevolution.piecewise_integrator_AD import pw_final_evolution_AD
from quocslib.utils.jax_utils import fidelity_funct as fidelity_funct_AD


def tensor_together(A):
    """Takes a list of matrices and multiplies them together with the tensor product"""
    res = np.kron(A[0], A[1])
    if len(A) > 2:
        for two in A[2:]:
            res = np.kron(res, two)
    else:
        res = res
    return res


def get_static_hamiltonian(nqu, J, g):
    """
    Get the static Hamiltonian for the Ising model
    :param nqu: Number of qubits
    :param J: Nearest neighbour coupling
    :param g: Next-nearest neighbour coupling
    :return: The Hamiltonian
    """
    dim = 2**nqu
    H0 = np.zeros((dim, dim), dtype=np.complex128)
    i2 = np.eye(2)
    sz = np.array([[1, 0], [0, -1]], dtype="complex")
    for j in range(nqu):
        # set up holding array
        rest = [i2] * nqu
        # set the correct elements to sz
        # check, so we can implement a loop around
        if j == nqu - 1:
            idx1 = j
            idx2 = 0
        else:
            idx1 = j
            idx2 = j + 1
        rest[idx1] = sz
        rest[idx2] = sz
        H0 = H0 - J * tensor_together(rest)

    for j in range(nqu):
        # set up holding array
        rest = [i2] * nqu
        # set the correct elements to sz
        # check, so we can implement a loop around
        if j == nqu - 1:
            idx1 = j
            idx2 = 1
        elif j == nqu - 2:
            idx1 = j
            idx2 = 0
        else:
            idx1 = j
            idx2 = j + 2
        rest[idx1] = sz
        rest[idx2] = sz
        H0 = H0 - g * tensor_together(rest)
    return H0


def get_control_hamiltonian(nqu: int):
    """"
    Get the control Hamiltonian for the Ising model
    :param nqu: Number of qubits
    :return: The Hamiltonian
    """
    dim = 2**nqu
    H_at_t = np.zeros((dim, dim), dtype=np.complex128)
    i2 = np.eye(2)
    sx = np.array([[0, 1], [1, 0]], dtype="complex")
    for j in range(nqu):
        # set up holding array
        rest = [i2] * nqu
        # set the correct elements to sz
        # check, so we can implement a loop around
        rest[j] = sx
        H_at_t = H_at_t + tensor_together(rest)
    return H_at_t


def get_initial_state(nqu: int):
    """
    Get the initial state for the Ising model
    :param nqu:
    :return: Initial state density matrix
    """
    rho0 = np.array([[1, 0], [0, 0]], dtype=np.complex128)
    state = [rho0] * nqu
    return tensor_together(state)


def get_target_state(nqu: int):
    """
    Get the target state for the Ising model
    :param nqu:
    :return: Target state density matrix
    """
    rhoT = np.array([[0, 0], [0, 1]], dtype=np.complex128)
    state = [rhoT] * nqu
    return tensor_together(state)


class IsingModel(AbstractFoM):
    """
    A figure of merit class for optimization of the problem defined by Alastair Marshall via
    https://arxiv.org/abs/2110.06187
    """
    def __init__(self, args_dict: dict = None):
        if args_dict is None:
            args_dict = {}

        ################################################################################################################
        # Dynamics variables
        ################################################################################################################
        self.n_qubits = args_dict.setdefault("n_qubits", 5)
        self.J = args_dict.setdefault("J", 1)
        self.g = args_dict.setdefault("g", 2)
        self.n_slices = args_dict.setdefault("n_slices", 100)

        self.H_drift = jnp.asarray(get_static_hamiltonian(self.n_qubits, self.J, self.g))
        self.H_control = jnp.asarray(get_control_hamiltonian(self.n_qubits))
        self.rho_0 = jnp.asarray(get_initial_state(self.n_qubits))
        self.rho_target = jnp.asarray(get_target_state(self.n_qubits))
        self.rho_final = jnp.asarray(np.zeros_like(self.rho_target))

    def get_control_Hamiltonians(self):
        return self.H_control

    def get_drift_Hamiltonian(self):
        return self.H_drift

    def get_target_state(self):
        return self.rho_target

    def get_initial_state(self):
        return self.rho_0

    def get_propagator(self,
                       pulses_list: list = jnp.array,
                       time_grids_list: list = jnp.array,
                       parameters_list: list = jnp.array) -> np.array:
        """ Compute and return the list of propagators """
        # jax.debug.print(pulses_list)
        # jax.debug.print("get_propagator, pulses_list: {}", pulses_list)
        drive = pulses_list[0, :].reshape(1, len(pulses_list[0, :]))
        n_slices = self.n_slices
        time_grid = time_grids_list[0, :]
        dt = time_grid[-1] / len(time_grid)
        # Compute the time evolution
        return pw_final_evolution_AD(drive, self.H_drift, [self.H_control], n_slices, dt, jnp.identity(2 ** self.n_qubits, dtype=np.complex128))

    def get_FoM(self,
                pulses: list = jnp.array,
                parameters: list = jnp.array,
                timegrids: list = jnp.array) -> dict:
        """
        Function to calculate the figure of merit from the pulses, parameters and timegrids.
        :param pulses: jnp.arrays of the pulses to be optimized.
        :param timegrids: jnp.arrays of the timegrids connected to the pulses.
        :param parameters: jnp.array of the parameters to be optimized.
        :return: dict - The figure of merit in a dictionary
        """
        U_final = self.get_propagator(pulses_list=pulses, time_grids_list=timegrids, parameters_list=parameters)
        rho_final = U_final @ self.rho_0 @ U_final.T.conj()
        fidelity = fidelity_funct_AD(rho_final, self.rho_target)
        return {"FoM": fidelity}



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import logging
import os
import sys
import time


def create_logger(results_path, date_time, create_logfile=True, console_info=True, is_debug=False):
    """
    Create a logger object based on the logging module. >It defines a custom format for the log messages, the date and
    printed messages and the log file name. The logger object is returned and can be used to log messages in the code.

    :param results_path: Path where the log file will be saved
    :param date_time: Date and time of the execution (for the log file name)
    :param create_logfile: Boolean to create the log file or not
    :param console_info: Boolean to show info in the console or not
    :param is_debug: Boolean to activate the debug mode
    :return: logger object
    """
    log_format = "%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s"
    date_format = "%d/%m/%Y %H:%M:%S"
    print_format = "%(levelname)-8s %(name)-12s %(message)s"
    log_filename = os.path.join(results_path, date_time + "_logging.log")
    log_debug_filename = os.path.join(results_path, date_time + "_logging_debug.log")

    logger = logging.getLogger("oc_logger")
    # Remove previous handlers if any
    logger.handlers = []
    # Default level for logger
    logger.setLevel(logging.INFO)
    # Console handler
    console_handler = logging.StreamHandler(sys.stdout)
    if console_info:
        console_handler.setLevel(logging.INFO)
    else:
        console_handler.setLevel(logging.ERROR)
    console_handler.setFormatter(logging.Formatter(print_format))
    # Log file handler
    if create_logfile:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(logging.INFO)
        file_handler.setFormatter(logging.Formatter(log_format, date_format))
        # Add handler for logfile to the logger
        logger.addHandler(file_handler)
    # Add handler for console to the logger
    logger.addHandler(console_handler)
    ################################################
    # Debug
    # In case of debug another object handler is activated
    ################################################
    if is_debug:
        # Log debug file handler
        debug_file_handler = logging.FileHandler(log_debug_filename)
        debug_file_handler.setLevel(logging.DEBUG)
        debug_file_handler.setFormatter(logging.Formatter(log_format, date_format))
        #
        logger.addHandler(debug_file_handler)
    return logger



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.parameters.BaseParameter import BaseParameter


class TimeParameter(BaseParameter):
    """
    Class for the time.
    Available in future updates:
    In case the user chooses to optimize the time, the code redefines the time in a function of a control parameter
    alpha:
    t = t_0 + alpha(t_max - t_min)
    where t_0 is the guess time, and t_max and t_min the limits for the varied time.
    The limits for the control parameter alpha are then defined as:
    alpha_in = 0
    alpha_max = (t_max - t_0)/(t_max - t_min)
    alpha_min = (t_min - t_0)/(t_max - t_min)
    alpha_variation = (amplitude_variation)/(t_max - t_min)
    """
    def __init__(
        self,
        map_index=-1,
        time_name="time",
        initial_value=1.0,
        is_optimization=False,
        lower_limit=0.0,
        upper_limit=1.0,
        amplitude_variation=0.1,
    ):
        """
        Constructor of the time parameter object.

        :param map_index: Index of the time in the map of parameters
        :param str time_name: Name of the time parameter assigned by the user
        :param float initial_value: Initial value or fixed value of the time
        :param bool is_optimization: True if the user wants also to optimize the time
        :param float lower_limit: Lower limit for the time
        :param float upper_limit: Upper limit for the time
        :param float amplitude_variation: Simplex points distance
        """
        # time name
        self.time_name = time_name
        self.is_optimization = is_optimization
        # Initialize the value of the time
        self.time = initial_value
        # Calculate the time variation
        self.time_variation = upper_limit - lower_limit
        if self.is_optimization:
            # Define the optimization coefficient alpha to be used in the optimization. Check the info for details
            max_variation = upper_limit - lower_limit
            alpha_max = (upper_limit - initial_value) / max_variation
            alpha_min = (lower_limit - initial_value) / max_variation
            alpha_in = 0.0
            alpha_variation = amplitude_variation / max_variation
            parameter_name = "alpha_" + time_name
            super().__init__(
                map_index=map_index,
                parameter_name=parameter_name,
                initial_value=alpha_in,
                lower_limit=alpha_min,
                upper_limit=alpha_max,
                amplitude_variation=alpha_variation,
            )
        # else:
        # Otherwise just call the parent constructor
        #    super().__init__(parameter_name=time_name, initial_value=initial_value)

    def get_time(self) -> float:
        """
        Returns the time. In case of time optimization the alpha parameter is used to get the optimized time

        :return float: time
        """
        if self.is_optimization:
            # value is the optimized parameter
            time = self.time + self.value * self.time_variation
        else:
            time = self.time
        return time



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from abc import ABCMeta, abstractmethod
import numpy as np


class AbstractFoM(metaclass=ABCMeta):
    """Abstract class for figure of merit evaluation"""

    def get_control_hamiltonians(self):
        """
        Method to get the control Hamiltonians. It is compulsory for gradient-based optimization.
        """
        raise NotImplementedError

    def get_propagator(self,
                       pulses_list: list = [],
                       time_grids_list: list = [],
                       parameters_list: list = []) -> np.array:
        """
        Method to get the propagator. It is compulsory for gradient-based optimization.

        :param list pulses_list: List of numpy arrays. Every array is a pulse.
        :param list time_grids_list: List of numpy arrays. Every array is a time grid that corresponds to a pulse.
        :param list parameters_list: List of floats for the parameters.
        """
        raise NotImplementedError

    def get_target_state(self):
        """
        Method to get the target state. It is compulsory for gradient-based optimization.
        """
        raise NotImplementedError

    def get_initial_state(self):
        """
        Method to get the target state. It is compulsory for gradient-based optimization.
        """
        raise NotImplementedError

    def get_drift_Hamiltonian(self):
        """
        Method to get the drift Hamiltonian. It is compulsory for gradient-based optimization.
        """
        raise NotImplementedError

    @abstractmethod
    def get_FoM(self,
                pulses_list: list = [],
                time_grids_list: list = [],
                parameters_list: list = []) -> dict:
        """
        Abstract method for figure of merit evaluation. It returns a dictionary with the FoM key inside.

        :param list pulses_list: List of numpy arrays. Every array is a pulse.
        :param list time_grids_list: List of numpy arrays. Every array is a time grid that corresponds to a pulse.
        :param list parameters_list: List of floats for the parameters.
        :return dict: The dictionary must contain at least the "FoM" key with the figure of merit as a float. It is
        advised to also provide an estimate or measured value for the standard deviation of the FoM under the key "std".
        Other possible keys can provide useful information for e.g. errors.
        """



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from quocslib.utils.dynamicimport import dynamic_import
from quocslib.pulses.BasePulse import BasePulse
from quocslib.tools.randomgenerator import RandomNumberGenerator


class ChoppedBasis(BasePulse):
    """
    General class for chopped basis. It inherits from the BasePulse class.
    """

    super_parameter_number: int

    def __init__(self, basis: dict = None, rng: RandomNumberGenerator = None, is_AD: bool = False, **kwargs):
        """
        Constructor of the ChoppedBasis class. It calls the constructor of the parent class BasePulse.

        :param dict basis: The dictionary of the basis is defined here.
        :param RandomNumberGenerator rng: The random number generator object.
        :param bool is_AD: Flag to activate the automatic differentiation.
        """
        super().__init__(rng=rng, is_AD=is_AD, **kwargs)
        super_parameter_distribution_dict = basis["random_super_parameter_distribution"]
        # Distribution attribute
        distribution_attribute = dynamic_import(
            attribute=super_parameter_distribution_dict.setdefault("distribution_attribute", None),
            module_name=super_parameter_distribution_dict.setdefault("distribution_module", None),
            class_name=super_parameter_distribution_dict.setdefault("distribution_class", None),
            name=super_parameter_distribution_dict.setdefault("distribution_name", None),
            class_type='superparameter_distribution')
        self.super_parameter_distribution_obj = distribution_attribute(self.super_parameter_number,
                                                                       super_parameter_distribution_dict,
                                                                       rng=self.rng)

    # Implement here other modules for Chopped Random Basis

    def update_chopped_basis(self) -> None:
        """
        Does nothing so far but can be overloaded to update the chopped basis.
        """



# a collection of potentially useful pulse tools
from scipy import interpolate
import matplotlib.pyplot as plt
import numpy as np


def visualise_pulse(dt, optimized_pulse):
    """
    Plots the pulses given in the optimized_pulse array.

    :param dt: time step
    :param optimized_pulse: array of optimized pulses
    :return: figure with the pulses over time
    """
    tArray = np.array([dt] * np.shape(optimized_pulse)[0]).cumsum() - dt
    f, ax = plt.subplots()
    for i in range(np.shape(optimized_pulse)[1]):
        ax.bar(tArray, optimized_pulse[:, i], label=str(i), width=dt / 2)
    ax.set_xlabel("Time")
    ax.legend()
    plt.show()
    return f


def resample_pulse(dt, optimized_pulse, new_time_axis):
    """
    Resamples the optimized pulse to a new time axis.

    :param dt: time step
    :param optimized_pulse: array of optimized pulses
    :param new_time_axis: new time axis
    :return: array of resampled pulses
    """
    ret = np.zeros((len(new_time_axis), np.shape(optimized_pulse)[1]))
    original_time = np.array([dt] * np.shape(optimized_pulse)[0]).cumsum() - dt
    for i in range(np.shape(optimized_pulse)[1]):
        e0 = optimized_pulse[0, i]
        e1 = optimized_pulse[-1, i]
        f = interpolate.interp1d(
            original_time,
            optimized_pulse[:, i],
            kind="nearest",
            fill_value=(e0, e1),
            bounds_error=False,
        )
        ret[:, i] = f(new_time_axis)
    return ret



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.handleexit.HandleExit import HandleExit
from quocslib.handleexit.AbstractHandleExit import AbstractHandleExit
from quocslib.optimizationalgorithms.OptimizationAlgorithm import OptimizationAlgorithm
from quocslib.utils.dynamicimport import dynamic_import
from quocslib.communication.AllInOneCommunication import AllInOneCommunication
from quocslib.utils.BestDump import BestDump
from quocslib.utils.AbstractFoM import AbstractFoM


class Optimizer:
    """
    The Optimizer class is the main class of the optimization. It is responsible for the initialization of the
    optimization algorithm and the communication object. The communication object is responsible for the
    communication between the optimization algorithm and the interface. The optimization algorithm is responsible
    for the optimization itself. The optimizer class is also responsible for the initialization of the
    HandleExit object if it is not configured and passed to the Optimizer object. The HandleExit object is
    responsible for the handling of the exit signals.
    It also has the method execute() to start the optimization and the method get_optimization_algorithm() to
    return the optimization algorithm object.
    """
    def __init__(self,
                 optimization_dict: dict = None,
                 FoM_object: AbstractFoM = None,
                 comm_signals_list: [list, list, list] = None,
                 handle_exit_obj: AbstractHandleExit = None):
        """
        Constructor of the Optimizer class. Initializes the objects for ExitHandling, Communication, Dumping and
        OptimizationAlgorithm.

        :param optimization_dict: Dictionary containing the optimization settings
        :param FoM_object: Figure of Merit object
        :param comm_signals_list: List of signals for the communication object
        :param handle_exit_obj: HandleExit object
        """
        # Handle exit
        if handle_exit_obj is None:
            handle_exit_obj = HandleExit()

        self.interface_job_name = optimization_dict.setdefault("optimization_client_name", "run")
        self.create_logfile = optimization_dict.setdefault("create_logfile", True)
        self.console_info = optimization_dict.setdefault("console_info", True)
        self.dump_format = optimization_dict.setdefault("dump_format", "npz")
        self.optimization_direction = optimization_dict["algorithm_settings"].setdefault("optimization_direction",
                                                                                         "minimization")
        self.communication_obj = AllInOneCommunication(interface_job_name=self.interface_job_name,
                                                       FoM_obj=FoM_object,
                                                       handle_exit_obj=handle_exit_obj,
                                                       dump_attribute=BestDump,
                                                       comm_signals_list=comm_signals_list,
                                                       create_logfile=self.create_logfile,
                                                       console_info=self.console_info,
                                                       dump_format=self.dump_format,
                                                       optimization_direction=self.optimization_direction)

        self.results_path = self.communication_obj.results_path

        algorithm_dict = optimization_dict['algorithm_settings']
        self.optimizer_attribute = dynamic_import(attribute=algorithm_dict.get("algorithm_attribute", None),
                                                  module_name=algorithm_dict.get("algorithm_module", None),
                                                  class_name=algorithm_dict.get("algorithm_class", None),
                                                  name=algorithm_dict.setdefault("algorithm_name", None),
                                                  class_type='algorithm')

        self.opt_alg_obj: OptimizationAlgorithm = self.optimizer_attribute(optimization_dict=optimization_dict,
                                                                           communication_obj=self.communication_obj,
                                                                           FoM_object=FoM_object)

    def execute(self):
        """
        Execute the optimization. It runs the begin(), run() and end() methods of the optimization algorithm.
        """
        self.opt_alg_obj.begin()
        self.opt_alg_obj.run()
        self.opt_alg_obj.end()

    def get_optimization_algorithm(self):
        """
        Returns the optimization algorithm object

        :return: Optimization algorithm object
        """
        return self.opt_alg_obj



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np
from scipy.linalg import expm


def pw_evolution(U_store, drive, A, B, n_slices, dt):
    """
    Computes the piecewise evolution of a system defined by the
    Hamiltonian H = A + drive * B and store the result in U_store

    :param List[np.matrix] U_store: The storage for all the computed propagators
    :param np.array drive: An array of dimension n_controls x n_slices that contains the amplitudes of the pulse
    :param np.matrix A: The drift Hamiltonian
    :param List[np.matrix] B: The control Hamiltonians in a list
    :param int n_slices: Number of slices
    :param float dt: The duration of each time slice
    :return List[np.matrix] U_store: The computed propagators
    """
    K = len(B)
    for i in range(n_slices):
        H = A
        for k in range(K):
            H = H + drive[k, i] * B[k]
        U_store[i] = expm(-1j * dt * H)
    return U_store


def pw_final_evolution(drive, A, B, n_slices, dt, U0):
    """
    Computes the piecewise evolution of a system defined by the
    Hamiltonian H = A + drive * B and concatenate all the propagators

    :param np.array drive: An array of dimension n_controls x n_slices that contains the amplitudes of the pulse
    :param np.matrix A: The drift Hamiltonian
    :param List[np.matrix] B: The control Hamiltonians in a list
    :param int n_slices: Number of slices
    :param float dt: The duration of each time slice
    :param np.matrix U0: The initial propagator to start from
    :return np.matrix: The final propagator
    """
    K = len(B)
    U = U0
    for i in range(n_slices):
        H = A
        for k in range(K):
            H = H + drive[k, i] * B[k]
        U = expm(-1j * dt * H) @ U
    return U



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import importlib
# import traceback
import os
from quocslib.utils.inputoutput import readjson

folder = os.path.dirname(os.path.realpath(__file__))
total_dict = readjson(os.path.join(folder, "map_dictionary.json"))


def dynamic_import(attribute=None,
                   module_name: str = None,
                   class_name: str = None,
                   name: str = None,
                   class_type: str = None) -> callable:
    """
    Function for dynamic import of object based on a module name, class name and type.

    :param attribute: The attribute of the class you want to use. It is an optional argument.
    :param module_name: Relative import of the module
    :param class_name: Name of the class inside the module
    :param name: Name of the class in the map_dictionary.json
    :param class_type: Type of class, i.e. algorithm, dsm_settings, basis or superparameter_distribution
    :return: The attribute to use to create the object
    """
    # If the attribute is already given, then just return the attribute
    if attribute is not None:
        return attribute

    elif name is not None:
        map_dict = {}
        try:
            if class_type == 'algorithm':
                map_dict = total_dict['opti_algorithm_map']
            elif class_type == 'dsm_settings':
                map_dict = total_dict['dsm_settings_map']
            elif class_type == 'basis':
                map_dict = total_dict['basis_map']
            elif class_type == 'superparameter_distribution':
                map_dict = total_dict['superparameter_distribution_map']
            elif class_type is None:
                raise Exception('The type of the class is not provided!')

            name_dict = {}

            if name in map_dict:
                name_dict = map_dict[name]
                module_name = name_dict["module_name"]
                class_name = name_dict["class_name"]

            attribute = getattr(importlib.import_module(module_name), class_name)
            return attribute

        # TODO Substitute the Exception with a proper Error
        except Exception as ex:
            # traceback.print_exc()
            print("{0}.py module does not exist or {1} is not the class in that module".format(module_name, class_name))
            return None

    elif all([module_name is not None, class_name is not None]):
        try:
            # provide backward - compatibility after renaming
            # module_name = module_name.replace(
            #     "quocslib.pulses.super_parameter.", "quocslib.pulses.superparameter."
            # )
            attribute = getattr(importlib.import_module(module_name), class_name)
            return attribute
        # TODO Substitute the Exception with a proper Error
        except Exception as ex:
            print("{0}.py module does not exist or {1} is not the class in that module".format(module_name, class_name))
            return None
    else:
        print("module_name: {0} and/or class_name: {1} are None".format(module_name, class_name))
        return None



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np

from quocslib.pulses.BasePulse import BasePulse
from quocslib.pulses.basis.ChoppedBasis import ChoppedBasis
from quocslib.tools.randomgenerator import RandomNumberGenerator


class DummyBasis(ChoppedBasis):
    """
    Dummy Basis class. It is used as a template for the creation of new basis.
    """
    amplitude_variation: float
    optimized_control_parameters: np.ndarray
    optimized_super_parameters: np.ndarray
    time_grid: np.ndarray

    def __init__(self, map_index: int, pulse_dictionary: dict, rng: RandomNumberGenerator = None, is_AD: bool = False):
        """
        Constructor of the Dummy Basis class. It is used as a template for the creation of new basis. It calls the
        constructor of the parent class ChoppedBasis.

        :param int map_index: Index number to use to get the control parameter.
        :param dict pulse_dictionary: The dictionary of the pulse is defined here.
        :param RandomNumberGenerator rng: Random number generator.
        :param bool is_AD: Flag to indicate if the pulse is used for the automatic differentiation.
        """
        #################
        # Basis dependent settings
        #################
        basis_dict = pulse_dictionary["basis"]
        # Super Parameter number i.e. the basis vector number in the pulse parametrization
        self.super_parameter_number = basis_dict.setdefault("basis_vector_number", 1)
        # Number of control parameters to be optimized
        self.control_parameters_number = 2 * self.super_parameter_number
        #################
        # Standard Basis Settings: amplitude limits, amplitude variation for the simplex,
        # distribution of super parameters, etc ...
        ################
        # Constructor of the parent classes, i.e. Base Pulse and Chopped Basis
        super().__init__(map_index=map_index, rng=rng, is_AD=is_AD, **pulse_dictionary)
        #################
        # Basis dependent settings
        #################
        # Scale coefficients: average distance of the points in the intial simplex
        self.scale_coefficients = (self.amplitude_variation / np.sqrt(2) * np.ones((self.control_parameters_number, )))
        # Initial value of the parameters in the pulse parametrization
        self.offset_coefficients = np.zeros((self.control_parameters_number, ))

    def _get_shaped_pulse(self) -> np.array:
        """
        Definition of the pulse parametrization. It is called at every function evaluation to build the pulse and
        return it as an array.

        :return np.array: The pulse as an array.
        """
        #################
        # Standard Basis Settings: amplitude limits, amplitude variation for the simplex,
        # distribution of super parameters, etc ...
        ################
        # Pulse initialization
        pulse = np.zeros(self.bins_number)
        # Final time definition
        final_time = self.final_time
        # Pulse creation
        xx = self.optimized_control_parameters
        w = self.super_parameter_distribution_obj.w
        t = self.time_grid
        #################
        # Basis dependent settings
        #################
        for ii in range(self.super_parameter_number):
            pulse += xx[2 * ii] * np.sin(2 * np.pi * w[ii] * t / final_time) + xx[2 * ii + 1] * np.cos(
                2 * np.pi * w[ii] * t / final_time)
        return pulse



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np

from quocslib.tools.randomgenerator import RandomNumberGenerator
from quocslib.pulses.BasePulse import BasePulse


class PiecewiseBasis(BasePulse):
    """
    General class for piecewise basis. Inherits from the BasePulse class.
    """
    amplitude_variation: float
    optimized_control_parameters: np.ndarray
    optimized_super_parameters: np.ndarray
    time_grid: np.ndarray

    def __init__(self, map_index, pulse_dictionary: dict, rng: RandomNumberGenerator = None, is_AD: bool = False):
        """
        Constructor of the PiecewiseBasis class. It calls the constructor of the parent class BasePulse.

        :param int map_index: Index number to use to get the control parameter.
        :param dict pulse_dictionary: The dictionary of the pulse is defined here.
        :param RandomNumberGenerator rng: The random number generator object.
        :param bool is_AD: Flag to indicate if the pulse is used in the AD optimization.
        """
        self.control_parameters_number = pulse_dictionary["bins_number"]
        super().__init__(map_index=map_index, rng=rng, is_AD=is_AD, **pulse_dictionary)
        #################
        # Basis dependent settings
        #################
        self.offset_coefficients = np.zeros((self.control_parameters_number, ))
        self.scale_coefficients = self.amplitude_variation * np.ones((self.control_parameters_number, ))

    def _get_shaped_pulse(self) -> np.array:
        """
        Definition of the pulse parametrization. It is called at every function evaluation to build the pulse and
        return it as an array.

        :return np.array: The pulse as an array.
        """
        #################
        # Standard Basis Settings: amplitude limits, amplitude variation for the simplex,
        # distribution of super parameters, etc ...
        ################
        pulse = self.optimized_control_parameters
        # self.debug_print("_get_shaped_pulse piecewisebasis, optimized_control_parameters: {}", pulse)
        return pulse



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import jax
import jax.scipy as jsp

from functools import partial


# def pw_evolution_AD_old(U_store, drive, A, B, n_slices, dt):
#     """Compute the piecewise evolution of a system defined by the
#     Hamiltonian H = A + drive * B and store the result in U_store
#
#     :param List[np.matrix] U_store: the storage for all of the computed propagators
#     :param np.array drive: an array of dimension n_controls x n_slices that contains the amplitudes of the pulse
#     :param np.matrix A: the drift Hamiltonian
#     :param List[np.matrix] B: the control Hamiltonians
#     :param int n_slices: number of slices
#     :param float dt: the duration of each time slice
#     :return None: Stores the new propagators so this doesn't return
#     """
#     K = len(B)
#     for i in range(n_slices):
#         H = A
#         for k in range(K):
#             H = H + drive[k, i] * B[k]
#         U_store.at[i].set(jsp.linalg.expm(-1j * dt * H))
#     return U_store


def pw_evolution_AD(U_store, drive, A, B, n_slices, dt):
    """
    Computes the piecewise evolution of a system defined by the
    Hamiltonian H = A + drive * B and store the result in U_store

    :param List[np.matrix] U_store: The storage for all the computed propagators
    :param np.array drive: An array of dimension n_controls x n_slices that contains the amplitudes of the pulse
    :param np.matrix A: The drift Hamiltonian
    :param List[np.matrix] B: The control Hamiltonians in a list
    :param int n_slices: Number of slices
    :param float dt: The duration of each time slice
    :return List[np.matrix] U_store: The computed propagators
    """
    K = len(B)
    for i in range(n_slices):
        H = A
        for k in range(K):
            H = H + drive[k, i] * B[k]
        U_store.at[i].set(jsp.linalg.expm(-1j * dt * H))
    return U_store


# def pw_final_evolution_AD_slow(drive, A, B, n_slices, dt, u0):
#     """Compute the piecewise evolution of a system defined by the
#     Hamiltonian H = A + drive * B and concatenate all the propagators
#
#     :param np.array drive: an array of dimension n_controls x n_slices that contains the amplitudes of the pulse
#     :param np.matrix A: the drift Hamiltonian
#     :param List[np.matrix] B: the control Hamiltonians
#     :param int n_slices: number of slices
#     :param np.matrix u0: the initial propagator to start from
#     :return np.matrix: the final propagator
#     """
#     K = len(B)
#     U = u0
#     for i in range(n_slices):
#         H = A
#         for k in range(K):
#             H = H + drive[k, i] * B[k]
#         U = jsp.linalg.expm(-1j * dt * H) @ U
#     return U


@partial(jax.jit, static_argnames=["n_slices"])
def pw_final_evolution_AD(drive, A, B, n_slices, dt, U0):
    """
    Computes the piecewise evolution of a system defined by the
    Hamiltonian H = A + drive * B and concatenate all the propagators

    :param np.array drive: An array of dimension n_controls x n_slices that contains the amplitudes of the pulse
    :param np.matrix A: The drift Hamiltonian
    :param List[np.matrix] B: The control Hamiltonians in a list
    :param int n_slices: Number of slices
    :param float dt: The duration of each time slice
    :param np.matrix U0: The initial propagator to start from
    :return np.matrix: The final propagator
    """
    U = U0

    def body_fun(i, val):
        K = len(B)
        H = A
        for k in range(K):
            H = H + drive[k, i] * B[k]
        Uint = jsp.linalg.expm(-1.0j * dt * H)
        return Uint @ val

    U = jax.lax.fori_loop(0, n_slices, body_fun, U)
    return U



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from abc import abstractmethod
import numpy as np
from datetime import datetime
import logging


class StoppingCriteria:
    """
    Parent class for the stopping criteria. It contains the common attributes and methods for all the stopping
    criteria.
    """

    def __init__(self, stopping_criteria: dict = None):
        """
        Constructor of the StoppingCriteria class. Sets the values from the stopping criteria dictionary or the
        default values if not provided.

        :param dict stopping_criteria: dictionary containing the settings for the stopping criteria
        """
        self.xatol = None
        self.fatol = None
        self.terminate_reason: str = ""
        self.is_converged = False
        self.terminate_reason = "Global Stopping Criterion reached."
        self.logger = logging.getLogger("oc_logger")
        self.max_eval = stopping_criteria.setdefault("max_eval", 10**10)
        self.time_lim = stopping_criteria.setdefault("time_lim", 10**10)
        self.direct_search_start_time = datetime.now()
        self.change_based_stop = stopping_criteria.setdefault("change_based_stop", {
            "cbs_funct_evals": 1,
            "cbs_change": 0
        })
        # if "cbs_funct_evals" not in self.change_based_stop:
        #     self.change_based_stop["cbs_funct_evals"] = 1
        self.curr_FoM_track: list = []

    @abstractmethod
    def check_stopping_criteria(self, **kwargs) -> None:
        """
        Abstract method for the stopping criteria. It must be implemented in the custom stopping criteria class.
        """
        raise ValueError("Must be implemented in the custom stopping criteria")

    # TODO Implement handle exit check
    def check_user_stop(self):
        if False:
            self.is_converged = True

    def check_func_eval_single_direct_search(self, func_evaluations_single_direct_search: int) -> [bool, str]:
        """
        Checks whether the maximum number of function evaluations has been exceeded

        :param float func_evaluations_single_direct_search: number of current function evaluation
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the termination
        """
        terminate_reason = "Exceeded number of allowed function evaluations per direct search."
        is_converged = False
        if func_evaluations_single_direct_search >= self.max_eval:
            is_converged = True
        return [is_converged, terminate_reason]

    def check_simplex_criterion(self, sim: np.array) -> [bool, str]:
        """
        Checks whether the simplex has converged

        :param np.array sim: current simplex
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        terminate_reason = "Convergence of the simplex."
        is_converged = False
        if np.max(np.ravel(np.abs(sim[1:] - sim[0]))) <= self.xatol:
            is_converged = True
        return [is_converged, terminate_reason]

    def check_f_size(self, fsim: np.array) -> [bool, str]:
        """
        Checks whether the FoM has converged inside this simplex

        :param np.array fsim: FoM values for current simplex
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the termination
        """
        terminate_reason = "Convergence of the FoM."
        is_converged = False

        if np.std(fsim) <= self.fatol:
            is_converged = True
            terminate_reason = "Convergence of the FoM (fatol)."
        return [is_converged, terminate_reason]

    def reset_direct_search_start_time(self) -> None:
        """
        Resets the direct_search_start_time for a new SI
        """
        self.direct_search_start_time = datetime.now()

    def check_direct_search_time_out(self) -> [bool, str]:
        """
        Checks whether the optimization in a direct search has been running for too long.

        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        end_time = datetime.now()
        minutes_diff = (end_time - self.direct_search_start_time).total_seconds() / 60.0
        terminate_reason = "Optimization time for direct search exceeded limit."
        is_converged = False
        if minutes_diff >= self.time_lim:
            is_converged = True
        return [is_converged, terminate_reason]

    def reset_curr_FoM_track_for_new_SI(self) -> None:
        """
        Resets the current FoM track list for a new SI
        """
        self.curr_FoM_track = []

    def add_to_FoM_track(self, FoM) -> None:
        """
        Adds new entry to current FoM track list

        :param float FoM: current FoM
        """
        if self.change_based_stop["cbs_funct_evals"] > 1:
            self.curr_FoM_track.append(FoM)

    def _check_cbs_stopping_crit(self) -> None:
        """
        Check the change-based stopping criterion
        """
        self.terminate_reason = "Change-based stopping criterion reached."
        cbs_funct_evals = self.change_based_stop["cbs_funct_evals"]
        cbs_change = self.change_based_stop["cbs_change"]
        if len(self.curr_FoM_track) >= cbs_funct_evals:
            last_foms = np.asarray(self.curr_FoM_track[-cbs_funct_evals:], dtype=np.float64)
            m, b = np.polyfit(range(cbs_funct_evals), last_foms, 1)
            current_change = abs(m * cbs_funct_evals)
            # check if the trend of changes is smaller than defined
            if current_change < cbs_change:
                self.is_converged = True

    def check_advanced_stopping_criteria(self) -> None:
        """
        Check the advanced stopping criteria if they are defined
        """
        try:
            if self.change_based_stop["cbs_funct_evals"] > 1:
                self._check_cbs_stopping_crit()
        except Exception as e:
            print(e)
            message = "Checking change-based stop failed! cbs_funct_evals not properly defined!"
            self.logger.error(message)



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import jax.scipy as jsp
import jax.numpy as jnp
from jax import jit

@jit
def state_to_dm(state: jnp.array):
    return jnp.outer(state, jnp.conj(state))

@jit
def _sqrtm(a):
    eig_vals, V = jnp.linalg.eigh(a)
    # eig_vals = jnp.linalg.eigvals(a)
    sqrt_eig_vals = jnp.sqrt(eig_vals)
    V_inv = jnp.linalg.inv(V)
    return jnp.matmul(jnp.matmul(V, jnp.diag(sqrt_eig_vals)), V_inv)

def _displace(a: jnp.ndarray, alpha: complex, offset=0) -> jnp.ndarray:
    """Displacement operator using jax

    Parameters
    ----------
    a : jnp.ndarray
        _description_
    alpha : complex
        _description_
    offset : int, optional
        _description_, by default 0

    Returns
    -------
    jnp.ndarray
        _description_
    """
    a_dag = jnp.transpose(jnp.conj(a))
    D = jsp.linalg.expm(alpha * a_dag - jnp.conj(alpha) * a)
    return D